"""

Tutorial provided as part of PyMP
M. Moussallam

"""
import os
import os.path as op
import matplotlib.pyplot as plt
import numpy as np
os.environ['PYMP_PATH'] = '/home/manu/workspace/PyMP/'
from PyMP.mdct import Dico
from PyMP import mp, Signal
from PyMP.tools.Misc import euclid_dist, hamming_dist

# Decomposing and visualizing the sparse dec
signal = Signal(op.join(os.environ['PYMP_PATH'],'data/Bach_prelude_4s.wav'), mono=True)

sig_occ1 = signal[:signal.length/2]
sig_occ2 = signal[signal.length/2:]

dico = Dico([128,1024,8192])
target_srr = 5
max_atom_num = 200
app_1, _ = mp.mp(sig_occ1, dico, target_srr, max_atom_num)
app_2, _ = mp.mp(sig_occ2, dico, target_srr, max_atom_num)

#plt.figure(figsize=(16,6))
#plt.subplot(121)
#app_1.plot_tf()
#plt.subplot(122)
#app_2.plot_tf()
#plt.show()


sp_vec_1 = app_1.to_array()[0]
sp_vec_2 = app_2.to_array()[0]

print "%1.5f, %1.5f"%(euclid_dist(sp_vec_1,sp_vec_2), hamming_dist(sp_vec_1,sp_vec_2)) 

# Now the information distance
from PyMP.mp_coder import joint_coding_distortion

# Measure the distortion of joint coding using approx of first patter as the reference
max_rate = 1000 # maximum bitrate allowed (in bits)
search_width = 1024 # maximum time shift allowed in samples
info_dist = joint_coding_distortion(sig_occ2, app_1, max_rate, search_width)
info_dist_rev = joint_coding_distortion(sig_occ1, app_2, max_rate, search_width)

print "%1.5f  - %1.5f"%(info_dist/target_srr, info_dist_rev/target_srr)


# building the similarity matrix
# Now load the long version
from PyMP.signals import LongSignal
seg_size = 5*8192
long_signal = LongSignal(op.join(os.environ['PYMP_PATH'],'data/Bach_prelude_40s.wav'),
                         seg_size,
                         mono=True, Noverlap=0.5)

# decomposing the long signal
apps, decays = mp.mp_long(long_signal,
               dico,
               target_srr, max_atom_num)




dists = np.zeros((long_signal.n_seg, len(apps)))    
mp._initialize_fftw(apps[0].dico, max_thread_num=1)
for idx in range(long_signal.n_seg):
    for jdx in range(idx):
        # test all preceeding segments only
        target_sig = long_signal.get_sub_signal(idx, 1, mono=True, pad=dico.get_pad(),fast_create=True)                                                                        
        dists[idx,jdx] = joint_coding_distortion(target_sig, apps[jdx],max_rate,1024, initfftw=False)  
                                                                                                                   
                                                                                                                   
mp._clean_fftw()    

plt.figure()
plt.imshow(dists)
plt.colorbar()
plt.show()





import numpy as np
from math import floor

from ...tools import Misc
from ... import win_server
from ... import log
from ... import parallelProjections
from ...baserand import AbstractSequenceBlock
from .. import block as mdct_block
from .. import atom as mdct_atom

# declare global win_server shared by all MDCT blocks instances
#global _PyServer
# _Logger
_PyServer = win_server.get_server()
_Logger = log.Log('SSMPBlocks', level=0)

class SequenceBlock(AbstractSequenceBlock, mdct_block.Block):
    """ block implementing the Randomized Pursuit

    Attributes
    ----------
    `sequence_type`: str
        The type of time-shift sequence, available choices are 
            *scale*
            *random*
            *gaussian*
            *binom*
            *dicho*
            *jump*
            *binary*
        default is *random* which use a uniform pseudo-random generator
    `shift_list`: array-like
        The actual sequence of subdictionary time-shifts
    `current_shift`: int
        The current time-shift
    `nbSim`: int
        Number of consecutive iterations with the same time-shift (default is 1)
    """

    # properties
    sequence_type = 'random'
    shift_list = []
    current_shift = 0
    currentSubF = 0 #DEPRECATED
    nb_consec_sim = 1
    w_long = None

    # constructor - initialize residual signal and projection matrix
    def __init__(self, length=0, resSignal=None, frameLen=0, 
                 randomType='random', nbSim=1, 
                 windowType=None,
                 seed=None):
        self.scale = length
        self.residual_signal = resSignal        
        super(SequenceBlock,self).__init__(randomType=randomType, nbSim=nbSim,                  
                                            seed=seed)

#            np.random.seed()

        if frameLen == 0:
            self.frame_len = length / 2
        else:
            self.frame_len = frameLen
        if self.residual_signal == None:
            raise ValueError("no signal given")

        self.framed_data_matrix = self.residual_signal.data
        self.frame_num = len(self.framed_data_matrix) / self.frame_len
        self.projs_matrix = np.zeros(len(self.framed_data_matrix))

        self.sequence_type = randomType
        if self.sequence_type == 'scale':
            self.shift_list = range(self.scale / 2)
        elif self.sequence_type == 'random':
            self.shift_list = [floor(
                (self.scale / 2) * (i - 0.5)) for i in np.random.random(self.scale)]
        elif self.sequence_type == 'gaussian':
            self.shift_list = [floor(self.scale / 8 * i)
                  for i in np.random.randn(self.scale / 2)]
            for k in self.shift_list:
                k = min(k, self.scale / 4)
                k = max(k, -self.scale / 4)

        elif self.sequence_type == 'binom':
            self.shift_list = Misc.binom(range(self.scale / 2))
        elif self.sequence_type == 'dicho':
            self.shift_list = Misc.dicho([], range(self.scale / 2))
        elif self.sequence_type == 'jump':
            self.shift_list = Misc.jump(range(self.scale / 2))
        elif self.sequence_type == 'sine':
            self.shift_list = Misc.sine(range(self.scale / 2))
#        elif self.randomType == 'triangle':
#            self.shift_list = Misc.triangle(range(self.scale/2))
        elif self.sequence_type == 'binary':
            self.shift_list = Misc.binary(range(self.scale / 2))

        else:
            self.shift_list = np.zeros(self.scale / 2)

        self.nb_consec_sim = nbSim
        self.windowType = windowType

    # The update method is nearly the same as CCBlock
    def update(self, new_res_signal, startFrameIdx=0, stopFrameIdx=-1, iteration_number=0):
        """ Same as superclass except that at each update, one need to pick a time shift from the sequence """

        # change the current Time-Shift if enough iterations have been done
        if (self.nb_consec_sim > 0):
            if (iteration_number % self.nb_consec_sim == 0):
                self.current_shift = self.shift_list[(
                    iteration_number / self.nb_consec_sim) % len(self.shift_list)]

        self.residual_signal = new_res_signal

        if stopFrameIdx < 0:
            endFrameIdx = self.frame_num - 1
        else:
            endFrameIdx = stopFrameIdx
        L = self.scale

        # update residual signal
        self.framed_data_matrix[startFrameIdx * L / 2: endFrameIdx * L / 2 + L] = self.residual_signal.data[startFrameIdx * self.frame_len: endFrameIdx * self.frame_len + 2 * self.frame_len]

        # TODO changes here
        self.compute_transform(startFrameIdx, stopFrameIdx)

        # TODO changes here
        self.find_max()

    # inner product computation through MCLT with all possible time shifts
    def compute_transform(self,   startingFrame=1, endFrame=-1):
        if self.w_long is None:
            self.initialize()

        if endFrame < 0:
            endFrame = self.frame_num - 2

        if startingFrame < 2:
            startingFrame = 2
#        L = self.scale
#        K = L/2;

        ############" changes  ##############
#        T = K/2 + self.current_shift
        # new version with C calculus
        parallelProjections.project(self.framed_data_matrix, self.best_score_tree,
                                                 self.projs_matrix,
                                                 self.locCoeff,
                                                 self.post_twid_vec,
                                                 startingFrame,
                                                 endFrame,
                                                 self.scale,
                                                 int(self.current_shift))

    def find_max(self):
        treeMaxIdx = self.best_score_tree.argmax()
        maxIdx = np.abs(self.projs_matrix[treeMaxIdx * self.scale /
            2: (treeMaxIdx + 1) * self.scale / 2]).argmax()

        self.maxIdx = maxIdx + treeMaxIdx * self.scale / 2
        self.max_value = self.projs_matrix[self.maxIdx]

    # construct the atom
    def get_max_atom(self):
        self.max_frame_idx = floor(self.maxIdx / (0.5 * self.scale))
        self.max_bin_idx = self.maxIdx - self.max_frame_idx * (0.5 * self.scale)
        Atom = mdct_atom.Atom(self.scale, 1, max((self.max_frame_idx * self.scale / 2) - self.scale / 4, 0), self.max_bin_idx, self.residual_signal.fs)
        Atom.frame = self.max_frame_idx
        Atom.mdct_value = self.max_value

        # new version : compute also its waveform through inverse MDCT
        Atom.waveform = self.synthesize_atom()

        if self.windowType == 'half1':
            Atom.waveform[0:self.scale / 2] = 0

        Atom.time_shift = self.current_shift
        Atom.time_position -= Atom.time_shift

        return Atom


class StochasticBlock(mdct_block.FullBlock):
    """ Block implementing the stochastic MP:
        atom selection is probabilistic in the full dictionary"""
        
    def __init__(self, length=0, resSignal=None, frameLen=0, 
                 randomType='proba',
                 windowType=None,
                 seed=None,
                 sigma = 1,
                 debug_level=0):
        _Logger.set_level(debug_level)
        
        self.scale = length
        self.residual_signal = resSignal
        self.seed = seed

        np.random.seed(self.seed)

        if frameLen == 0:
            self.frame_len = length / 2
        else:
            self.frame_len = frameLen
        if self.residual_signal == None:
            raise ValueError("no signal given")

        self.framed_data_matrix = self.residual_signal.data
        self.frame_num = len(self.framed_data_matrix) / self.frame_len
        
        self.projs_matrix = np.zeros((self.scale / 2, len(self.framed_data_matrix)))
        self.best_score_tree = np.zeros((self.scale / 2, self.frame_num))
        
#        for i in range(self.scale / 2):
##            self.projs_matrix[i] = zeros((self.scale/2 , self.scale/2))
#            self.projs_matrix.append( np.zeros(len(self.framed_data_matrix)))
#            self.best_score_tree[i] = np.zeros(self.frame_num)
#        self.projs_matrix = np.zeros(len(self.framed_data_matrix))

        self.randomType = randomType
        
        self.windowType = windowType
    
        self.sigma= sigma
    
    def compute_transform(self, startingFrame=1, endFrame=-1):
        if self.w_long is None:
            self.initialize()

        if endFrame < 0:
            endFrame = self.frame_num - 1

        if startingFrame < 1:
            startingFrame = 1

        startingFrame = 1
        endFrame = self.frame_num - 1

        L = self.scale
        K = L / 2
#        T = K/2
#        normaCoeffs = sqrt(2/float(K))
#        print startingFrame , endFrame
        for l in range(-K / 2, K / 2, 1):
            parallelProjections.project(self.framed_data_matrix,
                                                 self.best_score_tree[l + K / 2,:],
                                                 self.projs_matrix[
                                                     l + K / 2,:],
                                                 self.locCoeff,
                                                 self.post_twid_vec,
                                                 startingFrame,
                                                 endFrame,
                                                 self.scale, l)
    
    def find_max(self):
        """ the selection step is probabilistic: the max
            index is taken at random with a probability proportionnal to
            the projections 
            
            values are chosen with probability linearly proportionnal 
            to exp(sigma * projection)
            """        
        # draw a new index vector whose weights are poduct of
        # a random variable and the projection matrix
        probabilities = np.exp(self.sigma * np.abs(self.projs_matrix.ravel())) - np.exp(0)
        # normalize so it adds to 1                
        
        probabilities /= np.sum(probabilities)        
        # create the bins
        bins = np.add.accumulate(probabilities)
        
        # now draw an index at random: use it as the selected atom
        maxidx = np.digitize(np.random.random_sample(1), bins)[0]
        
#        print np.max(np.abs(self.projs_matrix))
        # deduce the value of the selected atom
        self.max_shift = maxidx / len(self.framed_data_matrix)
        self.maxIdx = maxidx % len(self.framed_data_matrix)
        
        self.max_value = self.projs_matrix[self.max_shift, self.maxIdx]
        _Logger.debug("Chose index %d - shift %d at random with value %1.5f"%(self.maxIdx,
                                                                              self.max_shift,
                                                                              self.max_value))
        
#        print self.max_value

    def get_max_atom(self):
        self.max_frame_idx = floor(self.maxIdx / (0.5 * self.scale))
        self.max_bin_idx = self.maxIdx - self.max_frame_idx * (0.5 * self.scale)
        Atom = mdct_atom.Atom(self.scale, 1,
                              max((self.max_frame_idx * self.scale / 2) - self.scale / 4, 0),
                              self.max_bin_idx, self.residual_signal.fs)
        Atom.frame = self.max_frame_idx
        Atom.mdct_value = self.max_value
        Atom.time_shift = -self.max_shift  + self.scale/4        

        Atom.time_position += Atom.time_shift
        # new version : compute also its waveform through inverse MDCT
        Atom.waveform = self.synthesize_atom()


        return Atom
    


#*****************************************************************************/
#                                                                            */
#                               mp_coder.py                                  */
#                                                                            */
#                        Matching Pursuit Library                            */
#                                                                            */
# M. Moussallam                                             Fri Nov 16 2012  */
# -------------------------------------------------------------------------- */


"""

Module mp_coder
===============
A collection of method handling the (theoretical) encoding of sparse approximations.



"""
import numpy as np
from . import approx
from mdct import atom as mdct_atom
from math import ceil, log, sqrt
from PyMP.parallelProjections import project_atom
from PyMP import mp

def simple_mdct_encoding(app,
                         target_bitrate,
                         Q=7,
                         encode_weights=True,
                         encode_indexes=True,
                         subsampling=1,
                         shift_penalty=False,
                         output_file_path=None,
                         output_all_indexes=False):
    """ Simple encoder of a sparse approximation

    Arguments:

        - `app`: a :class:`.Approx` object containing atoms from the decomposition

        - `target_bitrate`: a float indicating the target bitrate. Atoms are considered in decreasing amplitude order.
        The encoding will stop either when the target Bitrate it reached or when all atoms of `approx` have been considered

        - `Q`: The number of midtread quantizer steps. Default is 7, increase this number for higher bitrates

        - `shift_penalty`: a boolean indicating whether LOMP algorithm has been used, and addition time-shift parameters must be encoded

    Returns:

        - `snr`: The achieved Signal to Noise Ratio

        - `bitrate`: The achieved bitrate, not necessarily equal to the given target

        - `quantized_approx`: a :class:`.Approx` object containing the quantized atoms

    """
    # determine the fixed cost (in bits) of an atom's index
    index_cost = log(app.length * len(app.dico.sizes) / subsampling, 2)

    # determine the fixed cost (in bits) of an atom's weight
    weights_cost = Q

    # instantiate the quantized approximation
    quantized_approx = approx.Approx(app.dico,
                                    [],
                                    app.original_signal,
                                    app.length,
                                    app.fs)

    total = 0
    shift_cost = 0

    # First loop to evaluate the max of atom's weight and
    max_value = 0
    for atom in app.atoms:
        if abs(atom.mdct_value) > abs(max_value):
                max_value = atom.mdct_value

    # deduce the Quantizer step width
    quantizer_width = max_value / float(2 ** (Q - 1))

    if quantizer_width == 0:
        raise ValueError('zero found for quantizer step width...')

    if app.atom_number <= 0:
        raise ValueError('approx object has an empty list of atoms')

    # second loop: atom after atom
    for atom in app.atoms:

        value = atom.mdct_value

        # Quantize the atom's weight
        quantized_value = ceil(
            (value - quantizer_width / 2) / quantizer_width) * quantizer_width

        # If quantized value is 0, no need to include it
        if quantized_value == 0:
            # If one is 100% sure atom's weights are in decreasing order, we could stop right here
            # But in the general case, we're not...
            continue

        else:
            total += 1

        # instantiate the quantized atom
        quantized_atom = mdct_atom.Atom(atom.length,
                                       atom.amplitude,
                                       atom.time_position,
                                       atom.freq_bin,
                                       atom.fs,
                                       mdctCoeff=quantized_value)

        # recompute true waveform
        quantized_atom.frame = atom.frame
        quantized_atom.waveform = atom.waveform * (sqrt((
            quantized_atom.mdct_value ** 2) / (atom.mdct_value ** 2)))

        # add atom to quantized Approx
        quantized_approx.add(quantized_atom)

        # All time-shift optimal parameters live in an interval of length L/2 where
        # L is the atom scale
        if shift_penalty:
                shift_cost += log(atom.length / 2, 2)

        # estimate current bitrate: Each atom coding cost is fixed!!
        n_bits_sofar = (total * (index_cost + weights_cost)) + shift_cost
        br = float(n_bits_sofar) / (float(app.length) / float(app.fs))

        if br >= target_bitrate:
            break

    # Evaluate True Bitrate
    bitrate = float(n_bits_sofar) / (float(app.length) / float(app.fs))

    approx_energy = sum(quantized_approx.recomposed_signal.data ** 2)
    res_energy = sum(
        (app.original_signal.data - quantized_approx.recomposed_signal.data) ** 2)

    # Evaluate distorsion
    snr = 10.0 * log(approx_energy / res_energy, 10)

    return snr, bitrate, quantized_approx


def joint_coding_distortion(target_sig , ref_app, max_rate,
                          search_width , threshold=0 , doSubtract=True, 
                          debug=0 , discard = False,
                          precut=-1, initfftw=True):
    """ compute the joint coding distortion e.g. given a fixed maximum rate, 
    what is the distorsion achieved if coding the target_sig knowing the ref_app
    
    This is limited to a time adaptation of the atoms
    """    
    tolerances = [2]*len(ref_app.dico.sizes)
    # initialize the fftw    
    if initfftw:
        mp._initialize_fftw(ref_app.dico, max_thread_num=1)
    # initialize factored approx
    factorizedApprox = approx.Approx(ref_app.dico, 
                                       [], 
                                       target_sig, 
                                       ref_app.length, 
                                       ref_app.fs,
                                       fast_create=True)        
    timeShifts = np.zeros(ref_app.atom_number)
    atom_idx = 0
    rate = 0    
    residual = target_sig.data.copy()
    if debug > 0:
        print "starting factorization of " , ref_app.atom_number , " atoms"        
    while (rate < max_rate) and (atom_idx < ref_app.atom_number):
        # Stop only when the target rate is achieved or all atoms have been used        
        atom = ref_app[atom_idx]
        # make a copy of the atom
        newAtom = atom.copy()
        HalfWidth = (tolerances[ref_app.dico.sizes.index(atom.length)] -1) * atom.length/2;            
        
        # Search for best time shift using cross correlation        
        input1 = residual[atom.time_position - HalfWidth : atom.time_position + atom.length + HalfWidth]
        input2 = np.zeros((2*HalfWidth + atom.length))
        input2[HalfWidth:-HalfWidth] = atom.waveform        
        if not (input1.shape == input2.shape):
            raise ValueError("This certainly happens because you haven't sufficiently padded your signal") 
        scoreVec = np.array([0.0]);
        newts = project_atom(input1,input2 , scoreVec , atom.length)        
        score = scoreVec
    
        if debug>0:
            print "Score of " , score , " found"
        # handle MP atoms #
        if newAtom.time_shift is not None:
            newAtom.time_shift += newts
            newAtom.time_position += newts
        else:
            newAtom.time_shift = newts
            newAtom.time_position += newts
            atom.proj_score = atom.mdct_value
                        
        if debug>0:
            print "Factorizing with new offset: " , newts                
        
        if score <0:                    
            newAtom.waveform = -newAtom.waveform

        factorizedApprox.add(newAtom)
        if debug > 0:
            print "SRR Achieved of : " , factorizedApprox.compute_srr()            
        timeShifts[atom_idx] = newAtom.time_shift
            
        if doSubtract:                
            residual[newAtom.time_position : newAtom.time_position + newAtom.length ] -= newAtom.waveform
            
        rate += np.log2(abs(newts))+1
        if debug:
            print "Atom %d - rate of %1.3f"%(atom_idx, rate)
        atom_idx +=1
        
        # Use to prune the calculus
        if atom_idx>precut and precut>0:
            curdisto = factorizedApprox.compute_srr() 
            if curdisto<0:
                # pruning
                return curdisto
            else:
                # stop wasting time afterwards
                 precut = -1
                
    # cleaning
    if initfftw:
        mp._clean_fftw()

    # calculate achieved SNR :    
    return factorizedApprox.compute_srr()



'''
Abstract classes for Randomized Pursuits
'''
import numpy as np
from base import BaseDico, BaseBlock

class AbstractSequenceDico(BaseDico):
    """ RSSMP dictionary interface.
            
        Attributes:
        -----------        
        `sequence_type`: str
            The type of time-shift sequence, 
            available choices are *scale*,*random*,*gaussian*,*binom*,*dicho*,*jump*,*binary* 
            default is *random* which use a uniform pseudo-random generator
        `nb_consec_sim`: int
            Number of consecutive iterations with the same time-shift (default is 1)

    """

    # properties
    sequence_type = 'none'  # type of sequence , Scale , Random or Dicho
    it_num = 0  # memorizes the position in the sequence
    nb_consec_sim = 1
    # number of consecutive similar position
    nature = 'AbstractSequenceDico'

    # constructor
    def __init__(self, seq_type='random', nbSame=1, seed=None):
        self.sequence_type = seq_type        
        self.nb_consec_sim = nbSame            
        self.seed = seed
        
    def getSequences(self, length):
        sequences = []
        for block in self.blocks:
            sequences.append(block.shift_list[0:length])
        return sequences
    
    
class AbstractSequenceBlock(BaseBlock):
    """ block implementing RSSMP

    Attributes
    ----------
    `sequence_type`: str
        The type of time-shift sequence 
    `shift_list`: array-like
        The actual sequence of subdictionary time-shifts
    `current_shift`: int
        The current time-shift
    `nbSim`: int
        Number of consecutive iterations with the same time-shift (default is 1)
    """

    # properties
    sequence_type = 'random'
    shift_list = []
    current_shift = 0
    currentSubF = 0 #DEPRECATED
    nb_consec_sim = 1
    
    def __init__(self, 
                 randomType='random', nbSim=1,                  
                 seed=None):

        self.seed = seed
        np.random.seed(self.seed)
        
        
        


#
#
#                       PyMP.mdct.dico
#
#
#
# M. Moussallam                             Created on Nov 12, 2012
# -----------------------------------------------------------------------


"""
Module mdct.dico
================

This class inherits from :class:`.BaseDico` and is used to represent and manipulate multiscale MDCT dictionaries.
Dictionaries are mostly implemented as a collection of :class:`pymp_MDCTBlock` of various kind, according to the
nature of pursuit that is seeked.

This module describes 3 kind of blocks:
    - :class:`Dico`  is a Dico based on the standard MDCT transform. Which means atoms localizations
                 are constrained by the scale of the transform
                 The atom selected is simple the one that maximizes the correlation with the residual
                 it is directly indexes by the max absolute value of the MDCT bin
                 No further optimization of the selected atom is performed

    - :class:`LODico`   is a dictionary that performs a local optimization of the time localization of the selected atom

    - :class:`FullDico`  this object simulates a dictionary where atoms at all time localizations are available
                  This kind of dictionary can be thought of as a Toeplitz matrix
                  BEWARE: memory consumption and CPU load will be very high! only use with very small signals
                  (e.g. 1024 samples or so) Fairly intractable at higher dimensions
"""


import math
from xml.dom.minidom import Document, Element

from numpy import zeros, abs, sum, array

from .. import log
from ..base import BaseDico, BaseAtom
from . import block

#global _Logger
_Logger = log.Log('MDCTDico', level=0)


class Dico(BaseDico):
    """ class to handle multiscale MDCT dictionaries using MDCT blocks"""

    sizes = []
    tolerances = []
    nature = 'MDCT'
    blocks = None
    max_block_score = 0
    best_current_block = None
    starting_touched_index = 0
    ending_touched_index = -1

    # DEPRECATED
    forceHF = False
    use_c_optim = True

    def __init__(self, sizes=[], useC=True, forceHF=False, debug_level=None):
        if debug_level is not None:
            _Logger.set_level(debug_level)
        else:
            _Logger.set_level(0)
            
        self.sizes = sizes
        self.tolerances = [2]*len(self.sizes)
        self.use_c_optim = useC
        self.forceHF = forceHF
        self.touched_zone_width = max(self.sizes)
        _Logger.info('New dictionary created with sizes : ' + str(self.sizes))

    def find_block_by_scale(self, size):
        ''' Returns the index of the block corresponding to the given size or None if not found'''
        for i in range(len(self.sizes)):
            if size == self.sizes[i]:
                return i
        return None

    def initialize(self, residual_signal):
        ''' Create the collection of blocks specified by the MDCT sizes '''
        self.blocks = []
        self.best_current_block = None
        self.starting_touched_index = 0
        self.ending_touched_index = -1
        for mdctSize in self.sizes:
            self.blocks.append(block.Block(mdctSize,
                                           residual_signal, useC=self.use_c_optim, forceHF=self.forceHF))

    def init_proj_matrix(self, itNumbers):
        """ method used for monitoring the projection along iterations
            ONLY for DEV purposes"""
        projShape = self.blocks[0].projs_matrix.shape
        # Here all blocks have same number of projections
        return zeros((projShape[0] * len(self.sizes), itNumbers))

    def update_proj_matrix(self, projMatrix, iterationNumber, normedProj=False):
        """ method used for monitoring the projection along iterations
            ONLY for DEV purposes"""

        Lproj = self.blocks[0].projs_matrix.shape[0]
        for blockIdx in range(len(self.blocks)):
#            if normedProj:
# normCoeff = sqrt(sum(self.blocks[blockIdx].projectionMatrix**2))
#            else:
#            normCoeff = 1
            projMatrix[blockIdx * Lproj:(blockIdx + 1) * Lproj, iterationNumber] = self.blocks[
                blockIdx].projs_matrix  # /normCoeff

    def update(self, residualSignal, iteratioNumber=0, debug=0):
        ''' Update the projections in each block, only where it needs to be done as specified '''
        self.max_block_score = 0
        self.best_current_block = None

        if self.forceHF:
            self.maxHFBlockScore = 0
            self.bestCurrentHFBlock = None

        for block in self.blocks:
            startingTouchedFrame = int(
                math.floor(self.starting_touched_index / (block.scale / 2)))
            if self.ending_touched_index > 0:
                endingTouchedFrame = int(math.floor(self.ending_touched_index /
                                        (block.scale / 2))) + 1  # TODO check this
            else:
                endingTouchedFrame = -1
            _Logger.info("block: " + str(block.scale) + " : " +
                         str(startingTouchedFrame) + " " + str(endingTouchedFrame))

            block.update(
                residualSignal, startingTouchedFrame, endingTouchedFrame)

            if abs(block.max_value) > self.max_block_score:
                self.max_block_score = abs(block.max_value)
                self.best_current_block = block

    def get_best_atom(self, debug):
        if self.best_current_block == None:
            raise ValueError("no best block constructed, make sure inner product have been updated")

        if debug > 2:
            self.best_current_block.plot_proj_matrix()

        return self.best_current_block.get_max_atom()

    def compute_touched_zone(self, previousBestAtom, panic=False):
        ''' update zone computed from the previously selected atom '''
        self.starting_touched_index = previousBestAtom.time_position
        self.ending_touched_index = self.starting_touched_index + 1.5 * previousBestAtom.length
            
        # PANIC VERSION: REMOVE PRUNING OF PROJ COMPUTATIONS
        if panic:
            self.starting_touched_index = 0
            self.ending_touched_index = -1
            
#        self.starting_touched_index = previousBestAtom.time_position  - previousBestAtom.length/2
#        self.ending_touched_index = self.starting_touched_index + \
#            1.5 * previousBestAtom.length

#    def to_xml(self, doc):
#        ''' A routine to convert the dictionary to an XML node '''
#        if not isinstance(doc, Document):
#            raise TypeError('Xml document not provided')
#
#        DicoNode = doc.createElement('Dictionary')
#        DicoNode.setAttribute('nature', str(self.nature))
#        DicoNode.setAttribute('class', self.__class__.__name__)
#        SizesNode = doc.createElement('Sizes')
#        SizesNode.setAttribute('number', str(len(self.sizes)))
#        for size in self.sizes:
#            sizeNode = doc.createElement('Size')
#            sizeNode.setAttribute('scale', str(size))
#            SizesNode.appendChild(sizeNode)
#
#        DicoNode.appendChild(SizesNode)
#        return DicoNode

    def get_atom_key(self, atom, sigLength):
        ''' Get the atom index in the dictionary '''
        if not isinstance(atom, BaseAtom):
            return None

        block = [i for i in range(
            len(self.sizes)) if self.sizes[i] == atom.length][0]
        n = atom.time_position + 1
        frame = math.floor(float(n) / float(atom.length / 2)) + 1
        return int(block * sigLength + frame * float(atom.length / 2) + atom.freq_bin)

    #
    def get_projections(self, indexes, sigLength):
        """ additional method provided for Gradient Pursuits
            indexes formalism: key as in the py_pursuit_Approx Object :
            int(block*self.length +  frame*float(atom.length /2) + atom.frequencyBin)"""

        projections = []
        for index in indexes:
            block = int(math.floor(index / sigLength))
# frame = math.floor( (index - block*sigLength)   /  (self.sizes[block]  /2))
            projections.append(self.blocks[block].
                               projs_matrix[int(index - block * sigLength)])

        return projections


class LODico(Dico):
    """ Shift invariant MDCT dictionary
        Only difference is in the constructor and initialization: need to use LOBlocks """

    # properties
    HRsizes = []  # the sizes for which High resolution is seeked
    nature = 'LOMDCT'

    # constructor
    def __init__(self, sizes=[], hrsizes=None, useC=True, debug_level=None):
        ''' Basic contructor. By default all the block will have the locally optimized behvior but you
        can specify only a subset with the hrsizes variable '''
        if debug_level is not None:
            _Logger.set_level(debug_level)

        self.sizes = sizes
        self.use_c_optim = useC
        if hrsizes is not None:
            self.HRsizes = hrsizes
        else:
            self.HRsizes = sizes  # default behavior: compute high resolution for all scales

    def initialize(self, residual_signal):
        self.blocks = []
        self.best_current_block = None
        self.starting_touched_index = 0
        self.ending_touched_index = -1
        for mdctSize in self.sizes:
            # check whether this block should optimize time localization or not
            if mdctSize in self.HRsizes:
                self.blocks.append(block.LOBlock(mdctSize,
                                                 residual_signal, useC=self.use_c_optim))
            else:
                self.blocks.append(
                    block.Block(mdctSize, residual_signal, useC=self.use_c_optim))

    def get_projections(self, indexes, sigLength):
        """ additional method provided for Gradient Pursuits
            indexes formalism: key as in the py_pursuit_Approx Object :
            int(block*self.length +  frame*float(atom.length /2) + atom.frequencyBin)"""

        projections = []
        for index in indexes:
            block = int(math.floor(index / sigLength))
# frame = math.floor( (index - block*sigLength)   /  (self.sizes[block]  /2))
            value = self.blocks[block].projs_matrix[int(
                index - block * sigLength)]
            if value.real < 0:
                projections.append(-abs(value))
            else:
                projections.append(abs(value))

        return projections


class FullDico(Dico):
    """ This class handles blocks were MDCT products are computed for
        each and every possible time localization. Therefore Iterations
        are very slow but convergence should be optimum """
        # constructor
    nature = 'FullMDCT'

    def __init__(self, sizes=[]):
        self.sizes = sizes

    def initialize(self, residual_signal):
        self.blocks = []
        self.best_current_block = None
        self.starting_touched_index = 0
        self.ending_touched_index = -1
        for mdctSize in self.sizes:
            self.blocks.append(block.FullBlock(mdctSize, residual_signal))

    def init_proj_matrix(self, itNumbers):
        """ method used for monitoring the projection along iterations"""
        projShape = zeros(len(self.sizes))
        nbprojs = zeros(len(self.sizes))
        for bI in range(len(self.sizes)):
            projShape[bI] = (self.blocks[bI].projs_matrix[0].shape[0])
            nbprojs[bI] = (len(self.blocks[bI].projs_matrix))
            print bI, projShape[bI], nbprojs[bI]

        return zeros((sum(projShape * nbprojs), itNumbers))

    def update_proj_matrix(self, projMatrix, iterationNumber, normedProj=False):
        """ method used for monitoring the projection along iterations"""
        ide = 0
        for blockIdx in range(len(self.blocks)):
            projMat = array(
                self.blocks[blockIdx].projs_matrix.values()).flatten()
#            if normedProj:
#                normCoeff = sqrt(sum(projMat**2))
#            else:
#                normCoeff = 1
# print projMat.shape , projMatrix[id:id+projMat.shape[0],
# iterationNumber].shape
            projMatrix[ide:ide + projMat.shape[0],
                       iterationNumber] = projMat  # /normCoeff
            ide += projMat.shape[0]

class SpreadDico(Dico):
    # !!!!!!!!!!! UNDER DEVELOPPMENT DO NOT USE
    def __init__(self , sizes=[] ,nature = 'SpreadMDCT' ,debug_level=None ,
                 all_scales = True , spread_scales = [], penalty=0.5,
                 maskSize = 1, 
                 mask_time = -1, mask_freq = 1):
        # TODO implement frequency masking settings
        
        if debug_level is not None:
            _Logger.set_level(debug_level)
        
        self.sizes = sizes
        self.nature = nature
        self.penalty = penalty
        self.maskSize = maskSize
        
        if mask_time > 0:
            self.mask_times = max(sizes)*mask_time/array(sizes)
        else:
            self.mask_times = [self.maskSize]*len(sizes)
        
        self.mask_freq = mask_freq
        
        if all_scales:
            self.spreadScales = self.sizes
        else:
            self.spreadScales = spread_scales

        _Logger.info('New dictionary created with sizes : ' +str(self.sizes))

    def initialize(self , res_signal):
        """ Initialize As many blocks as number of signals x number of window sizes
        """
        self.blocks = []
        self.best_current_block = None
        self.starting_touched_index = 0
        self.ending_touched_index = -1
        for mdctSize in self.sizes:
            if mdctSize in self.spreadScales:
                self.blocks.append(block.SpreadBlock(
                    mdctSize , res_signal))
            else:
                self.blocks.append(block.Block(
                    mdctSize , res_signal))
    
    def get_best_atom(self, debug):
        if self.best_current_block == None:
            raise ValueError("no best block constructed, make sure inner product have been updated")

        if debug > 2:
            self.best_current_block.plot_proj_matrix()

        best_atom = self.best_current_block.get_max_atom()
        for blocki in range(len(self.blocks)):
            if isinstance(self.blocks[blocki], block.SpreadBlock):
                self.blocks[blocki].compute_mask(best_atom,                                                 
                                                 self.mask_freq,
                                                 self.mask_times[blocki],
                                                 self.penalty)

        return best_atom
#def fromXml(xmlNode):
#    ''' Export routine NOT FULLY TESTED '''
#    if not isinstance(xmlNode, Element):
#        raise TypeError('Xml element not provided')
#
#    # retrieve sizes
#    for e in xmlNode.childNodes:
#        if e.localName == 'Sizes':
#            sizesNode = e
#            break
#
##    sizesNode = xmlNode.childNodes[0]
#    sizes = []
#    for node in sizesNode.childNodes:
#        if node.localName == 'Size':
#            sizes.append(int(node.getAttribute('scale')))
#
#    if xmlNode.getAttribute('class') == 'Dico':
#        return Dico(sizes)
#
#    elif xmlNode.getAttribute('class') == 'LODico':
#        return LODico(sizes)
#
#    elif xmlNode.getAttribute('class') == 'FullDico':
#        return FullDico(sizes)



from math import floor


def dicho(x, sequence):
    """ creates a dichotomic sequence """
    if len(sequence) < 1:
        return x

    if len(sequence) > 2:
        x.append(sequence[len(sequence) / 2])
        dicho(x, sequence[1:len(sequence) / 2])
        dicho(x, sequence[len(sequence) / 2 + 1:-1])

    x.append(sequence[-1])
    x.append(sequence[0])
    return x


def jump(sequence):
    L = len(sequence)
    x = []
    for l in range(L):
        x.append(sequence[((l * ((L / 2) + 1)) % L)])
    return x


def sine(sequence):
    from numpy import sin, pi
    L = len(sequence)
    x = [abs(int((L / 2) * sin(x))) for x in sequence]
    return x


def binary(sequence):
    L = len(sequence)
    x = (sequence[0], sequence[L / 2]) * int(L / 2)
    return x


def binom(sequence):
    "uses a bernoulli variable to generate a white noise by inverse fft"
    L = len(sequence)
    from numpy.random import binomial

    p = 0.5
    binom = binomial(1, p, size=L)

    from numpy.fft import ifft
    from numpy import real
    from math import sqrt, pi
    from cmath import exp
    randPhases = [(L / 8) * exp(-1j * pi * i) for i in binom]
    x = sqrt(L) * real(ifft(randPhases, L))

    for i in range(len(x)):
        x[i] = floor(x[i])

    return x


def euclid_dist(x, y):
    from numpy import log10, linalg 
    return log10(linalg.norm(x - y)/(linalg.norm(x)*linalg.norm(y)))

def hamming_dist(x,y):
    x_ind = x.nonzero()[0]
    y_ind = y.nonzero()[0]
    import numpy as np
    same = float(np.count_nonzero(np.intersect1d(x_ind,y_ind)))
    all = float(np.count_nonzero(np.union1d(x_ind,y_ind)))    
    return 1.0-(same/all)

###############" testing ##########"
# print jump(range(16))
# print sine(range(64))
# print binary(range(16))



# This Python file uses the following encoding: utf-8

#*****************************************************************************/
#                                                                            */
#                               mp.py                                        */
#                                                                            */
#                        Matching Pursuit Library                            */
#                                                                            */
# M. Moussallam                                             Mon Aug 16 2010  */
# -------------------------------------------------------------------------- */


'''
Module mp
=========

Greedy algorithms using the pymp objects
----------------------------------------

'''
import math
import numpy as np
# these imports are for debugging purposes
import os.path

# PyMP object imports
import PyMP.approx as Approx
from PyMP.base import BaseDico as Dico
import PyMP.signals as Signal
import PyMP.log as log
import PyMP.win_server as win_server

# this import is needed for using the fast projection written in C extension
# module
try:
    import PyMP.parallelProjections as parproj
except ImportError:
    print ''' Failed to load the parallelProjections extension module'''

_PyServer = win_server.get_server()
_Logger = log.Log('mp', noContext=True)


def mp(orig_signal,
        dictionary,
        target_srr,
        max_it_num,
        debug=0,
        pad=True,
        clean=False,
        silent_fail=False,
        debug_iteration=-1,
        unpad=False,
        max_thread_num=None):
    """Matching Pursuit Decomposition    

    This is the original, plain Matching Pursuit Algorithm

    Parameters
    ----------
    orig_signal :  Signal
        the original signal (as a :class:`.Signal` object)  to decompose
    dictionary : BaseDico
        the dictionary (as a :class:`.Dico` object)  on which to decompose `orig_signal`
    target_srr : float
        a target Signal to Residual Ratio (SRR)
    max_it_num : int
        the maximum number of iteration allowed
    debug : int, optional
        requested debug level, default is 0
    pad : bool, optional
        whether to pad the signal's edges with zeroes
            
        
    Returns
    -------
    approx : Approx
        A :class:`.Approx` object encapsulating the approximant
    decay : list
        A list of residual's energy across iterations

    Other Parameters
    ----------------
    clean : bool, optional
        whether perform on the fly cleaning of atom's waveforms to 
        limit memory consumption
    silent_fail : bool, optional
        whether to allow energy increase when subtracting atoms
    debug_iteration : int, optional
        specify a specific iteration number at which the debug level 
        should be raised to maximum
    unpad : bool, optional
        whether to remove added edges zeroes after decomposition
    max_thread_num : int, optional
        the maximum number of threads allocated

    Raises
    ------
    ValueError
        If a selected atom causes the residual signal's energy to increase, see :func:`.Signal.subtract`

    Examples
    --------
    For decomposing the signal `x` on the Dictionary D at either SRR of 10 dB or using 1000 atoms:
    x must be a :class:`.Signal` and D a :class:`.Dico` object
    
    >>> approx, decay = mp.mp(x, D, 10, 1000)

    

    See Also
    --------
    greedy : a decomposition method where the update strategy (MP, OMP, GP or CMP)
        can be set as a parameter

    Notes
    -----
    This is the original Matching Pursuit Algorithm transcribed to 
    work with PyMP objects encapsulating signals :math:`x` and dictionaries :math:`\Phi`
    It builds an approximant :math:`\hat{x}` of :math:`x`
    as a linear combination of atoms from :math:`\Phi`
    
    """

    # back compatibility - use debug levels now
    if debug is not None:
        _Logger.set_level(debug)

    # CHECKING INPUTS
    if not isinstance(orig_signal, Signal.Signal):
        raise TypeError("MP will only accept objects inheriting from PyMP.signals.Signal as input signals")

    if not isinstance(dictionary, Dico):
        raise TypeError("MP will only accept objects inheriting from PyMP.base.BaseDico as dictionaries")

    # optional add zeroes to the edge
    if pad:
        orig_signal.pad(dictionary.get_pad())
    res_signal = orig_signal.copy()

    # FFTW Optimization for C code: initialize module global variables
    _initialize_fftw(dictionary, max_thread_num)

    # initialize blocks
    dictionary.initialize(res_signal)

    # initialize approximant
    current_approx = Approx.Approx(
        dictionary, [], orig_signal, debug_level=debug)

    # residualEnergy
    res_energy_list = []

    it_number = 0
    current_srr = current_approx.compute_srr()

    # check if signal has null energy
    if res_signal.energy == 0:
        raise ValueError(" Null signal energy ")

    res_energy_list.append(res_signal.energy)

    # Decomposition loop: stopping criteria is either SNR or iteration number
    while (current_srr < target_srr) & (it_number < max_it_num):

        if (it_number == debug_iteration):
            debug = 3
            _Logger.set_level(3)

        best_atom = _mp_loop(dictionary, debug,
                             silent_fail, unpad, res_signal,
                             current_approx,
                             res_energy_list, it_number)

        # compute new SRR and increment iteration Number
        current_srr = current_approx.compute_srr(res_signal)

        _Logger.debug("SRR reached of " + str(current_srr) +
                      " at iteration " + str(it_number))

        it_number += 1

        # cleaning for memory consumption control
        if clean:
            del best_atom.waveform

    # VERY IMPORTANT CLEANING STAGE!
    _clean_fftw()

    return current_approx, res_energy_list


def mp_continue(current_approx,
                orig_signal,
                dictionary,
                target_srr,
                max_it_num,
                debug=0,
                pad=True,
                silent_fail=False,
                unpad=False):
    """ routine that restarts a decomposition from an existing,
    incomplete approximation """

    if not isinstance(current_approx, Approx.Approx):
        raise TypeError("provided object is not a py_pursuit_Approx")

    # optional add zeroes to the edge
    if pad:
        orig_signal.pad(dictionary.get_pad())

    res_signal = orig_signal.copy()

    # retrieve approximation from residual
    if current_approx.recomposed_signal is not None:
        res_signal.data -= current_approx.recomposed_signal.data
        res_signal.energy = sum(res_signal.data ** 2)
    else:
        for atom in current_approx.atoms:
            res_signal.subtract(atom, debug)

    _initialize_fftw(dictionary)

    # initialize blocks
    dictionary.initialize(res_signal)

    # residualEnergy
    res_energy_list = [res_signal.energy]

    it_num = 0
    current_srr = current_approx.compute_srr()

    while (current_srr < target_srr) & (it_num < max_it_num):

        # lauches a mp loop: update projection,
        # retrieve best atom and subtract from residual

        _mp_loop(dictionary, debug,
                 silent_fail, unpad,
                 res_signal, current_approx,
                 res_energy_list, it_num)

        # compute new SRR and increment iteration Number
        current_srr = current_approx.compute_srr(res_signal)
        it_num += 1

        if debug > 0:
            print "SRR reached of ", current_srr, " at iteration ", it_num

    # VERY IMPORTANT CLEANING STAGE!
    _clean_fftw()

    return current_approx, res_energy_list


def mp_long(orig_longsignal,
            dictionary,
            target_srr=10,
            max_it_num=100,
            debug=False,
            pad=True,
            output_dir='',
            write=False):
    """NOT FULLY TESTED treats long signals , return a collection of approximants ,
        one by segment
        You can use the FusionApprox method of the Approx class to recover a
        single approx object from this collection"""

    output_dir = output_dir + str(target_srr) + '/'

    if not os.path.exists(output_dir) and write:
        os.makedirs(output_dir)

    # test inputs
    if not isinstance(orig_longsignal, Signal.LongSignal):
        raise TypeError(
            'Signal is not a LongSignal object ')

    if not isinstance(dictionary, Dico):
        raise TypeError('Dico is not a Dico object')

    Nsegments = int(orig_longsignal.n_seg)

    # initialize output
    approximants = []
    decays = []
    # Loop on segments
    for segIdx in range(Nsegments):
        if debug > 0:
            print 'Starting work on segment ' + str(
                segIdx) + ' over ' + str(Nsegments)

        subSignal = orig_longsignal.get_sub_signal(
            segIdx, 1, True, False, 0, 0)
        approx, decay = mp(subSignal, dictionary, target_srr,
                           max_it_num, debug=debug, pad=pad)
        approximants.append(approx)
        decays.append(decay)

        # save the output in xml formatting
        if write:
            approximants[segIdx].write_to_xml('Srr' + str(target_srr) +
                                              '_Seg' + str(segIdx) +
                                              '_Over_' + str(Nsegments) +
                                              '.xml', output_dir)

    return approximants, decays


def locomp(original_signal,
           dictionary,
           target_srr,
           max_it_num,
           debug=0,
           pad=True,
           clean=False,
           silent_fail=False,
           unpad=False,
           approximate=False):
    """
    # EXPERIMENTAL: NOT FULLY TESTED Yet
    # Orthogonal Matching Pursuit and Gradient Pursuit Loop

    We try try to implement the Localized OMP/GP described by Mailhe et al in:
    Mailhé, B., Gribonval, R., Vandergheynst, P., & Bimbot, F. (2011).
    Fast orthogonal sparse approximation algorithms over local dictionaries.
    published in Signal Processing.

    This means that least square (for OMP) or gradient descent (for GP)
     is only performed on a local subdictionary
    in order to accelerate computation

    parameters
    ----------

    approximate : if True, perform Gradient Pursuit instead of locOMP

    outputs:
    -------

    same as mp

    """

    # back compatibility - use debug levels now
    if debug is not None:
        _Logger.set_level(debug)

    # optional add zeroes to the edge
    if pad:
        original_signal.pad(dictionary.get_pad())
    res_signal = original_signal.copy()

    # FFTW C code optimization
    _initialize_fftw(dictionary)

    # initialize blocks
    dictionary.initialize(res_signal)

    # initialize approximant
    current_approx = Approx.Approx(dictionary, [], original_signal)

    # residualEnergy
    res_energy_list = []

    it_number = 0
    current_srr = current_approx.compute_srr()

    # loop is same as mp except for orthogonal projection of the atom on the
    # residual
    # Decomposition loop: stopping criteria is either SNR or iteration number
    # check if signal has null energy
    if res_signal.energy == 0:
        raise ValueError(" Null signal energy ")

    res_energy_list.append(res_signal.energy)

    while (current_srr < target_srr) & (it_number < max_it_num):

        if not approximate:
            best_atom = _locomp_loop(dictionary, debug,
                                     silent_fail, unpad, res_signal,
                                     current_approx,
                                     res_energy_list, it_number)
        else:
            best_atom = _locgp_loop(dictionary, debug,
                                    silent_fail, unpad, res_signal,
                                    current_approx,
                                    res_energy_list, it_number)

        # compute new SRR and increment iteration Number
        current_srr = current_approx.compute_srr(res_signal)

        _Logger.debug("SRR reached of " + str(current_srr) +
                      " at iteration " + str(it_number))

        it_number += 1

        # cleaning for memory consumption control
        if clean:
            del best_atom.waveform

    # VERY IMPORTANT CLEANING STAGE!
    _clean_fftw()

    return current_approx, res_energy_list


def greedy(orig_signal,
           dictionary,
           target_srr,
           max_it_num,
           debug=0,
           pad=True,
           clean=False,
           silent_fail=False,
           debug_iteration=-1,
           unpad=False,
           update='mp',
           max_thread_num=None):

    """Common Greedy Pursuit Loop Options are detailed below:

    Parameters
    ----------
    orig_signal :  Signal
        the original signal (as a :class:`.Signal` object)  to decompose
    dictionary : BaseDico
        the dictionary (as a :class:`.Dico` object)  on which to decompose `orig_signal`
    target_srr : float
        a target Signal to Residual Ratio (SRR)
    max_it_num : int
        the maximum number of iteration allowed
    debug : int, optional
        requested debug level, default is 0
    pad : bool, optional
        whether to pad the signal's edges with zeroes
    update : {'mp', 'omp', 'locomp', 'locgp'}, optional
        choice of update loop to consider
            
        
    Returns
    -------
    approx : Approx
        A :class:`.Approx` object encapsulating the approximant
    decay : list
        A list of residual's energy across iterations

    Other Parameters
    ----------------
    clean : bool, optional
        whether perform on the fly cleaning of atom's waveforms to 
        limit memory consumption
    silent_fail : bool, optional
        whether to allow energy increase when subtracting atoms
    debug_iteration : int, optional
        specify a specific iteration number at which the debug level 
        should be raised to maximum
    unpad : bool, optional
        whether to remove added edges zeroes after decomposition
    max_thread_num : int, optional
        the maximum number of threads allocated

    Raises
    ------
    ValueError
        If a selected atom causes the residual signal's energy to increase, see :func:`.Signal.subtract`


    Examples
    --------
    For decomposing the signal `x` on the Dictionary D at either SRR of 10 dB or using 1000 atoms:
    x must be a :class:`.Signal` and D a :class:`.Dico` object
    
    >>> approx, decay = mp.greedy(x, D, 10, 1000, 'mp')

    

    See Also
    --------
    mp : a decomposition method with plain mp only        

    """

    # back compatibility - use debug levels now
    if debug is not None:
        _Logger.set_level(debug)

    # CHECKING INPUTS
    if not isinstance(orig_signal, Signal.Signal):
        raise TypeError("MP will only accept objects inheriting from PyMP.signals.Signal as input signals")

    if not isinstance(dictionary, Dico):
        raise TypeError("MP will only accept objects inheriting from PyMP.base.BaseDico as dictionaries")

    # optional add zeroes to the edge
    if pad:
        orig_signal.pad(dictionary.get_pad())
    res_signal = orig_signal.copy()

    # FFTW Optimization for C code: initialize module global variables
    _initialize_fftw(dictionary, max_thread_num=max_thread_num)

    # initialize blocks
    dictionary.initialize(res_signal)

    # initialize approximant
    current_approx = Approx.Approx(
        dictionary, [], orig_signal, debug_level=debug)

    # residualEnergy
    res_energy_list = []

    it_number = 0
    current_srr = current_approx.compute_srr()

    # check if signal has null energy
    if res_signal.energy == 0:
        raise ValueError(" Null signal energy ")

    res_energy_list.append(res_signal.energy)

    if update not in ['mp', 'locomp', 'locgp', 'omp']:
        raise ValueError('Unrecognized update rule')

    # Decomposition loop: stopping criteria is either SNR or iteration number
    while (current_srr < target_srr) & (it_number < max_it_num):

        if (it_number == debug_iteration):
            debug = 3
            _Logger.set_level(3)

        if update == 'mp':
            best_atom = _mp_loop(dictionary, debug,
                                 silent_fail, unpad, res_signal,
                                 current_approx,
                                 res_energy_list, it_number)
        elif update == 'locomp':
            best_atom = _locomp_loop(dictionary, debug,
                                     silent_fail, unpad, res_signal,
                                     current_approx,
                                     res_energy_list, it_number)
        elif update == 'omp':
            best_atom = _omp_loop(dictionary, debug,
                                  silent_fail, unpad, res_signal,
                                  current_approx,
                                  res_energy_list, it_number)
        if update == 'locgp':
            best_atom = _locgp_loop(dictionary, debug,
                                    silent_fail, unpad, res_signal,
                                    current_approx,
                                    res_energy_list, it_number)

        # compute new SRR and increment iteration Number
        current_srr = current_approx.compute_srr(res_signal)

        _Logger.debug("SRR reached of " + str(current_srr) +
                      " at iteration " + str(it_number))

        it_number += 1

        # cleaning for memory consumption control
        if clean:
            del best_atom.waveform

    # VERY IMPORTANT CLEANING STAGE!
    _clean_fftw()

    return current_approx, res_energy_list


def _itprint_(it_num, best_atom):
    """ debug printing"""
    strO = ("It: " + str(it_num) + " Selected atom" + str(best_atom))
#    print strO
    return strO


def _plot_proj(best_atom, residual_sig):
    import matplotlib.pyplot as plt
    plt.figure()
    plt.plot(residual_sig.data[best_atom.
                               time_position: best_atom.time_position + best_atom.length])
    plt.plot(best_atom.waveform)
    plt.plot(residual_sig.data[best_atom.time_position: best_atom.time_position +
                               best_atom.length] - best_atom.waveform, ':')
    plt.legend(('Signal', 'Atom substracted', 'residual'))
    plt.show()


def _mp_loop(dictionary, debug, silent_fail,
             unpad, res_signal, current_approx,
             res_energy, it_number):
    #===========================================================================
    # # Internal MP loop
    #=========================================================================    
    # Compute inner products and selects the best atom
    dictionary.update(res_signal, it_number)
    # retrieve the best correlated atom
    best_atom = dictionary.get_best_atom(debug)
    if best_atom is None:
        print 'No atom selected anymore'
#            return current_approx, res_energy
    if debug > 0:
        _Logger.debug(_itprint_(it_number, best_atom))
    try:
        res_signal.subtract(best_atom, debug)
        dictionary.compute_touched_zone(best_atom)
    except ValueError:
        if not silent_fail:
            _Logger.error("Something wrong happened at iteration " +
                          str(it_number) + " atom substraction abandonned")
            print "Atom Selected: ", best_atom
            print res_signal.length, best_atom.time_position, best_atom.length            
            if debug > 1:
                import matplotlib.pyplot as plt
                plt.figure()
                plt.plot(res_signal.data[best_atom.time_position:
                         best_atom.time_position + best_atom.length],'o-')
                plt.plot(best_atom.waveform,'x--')
                plt.plot(res_signal.data[best_atom.time_position:
                                         best_atom.time_position + best_atom.length] - best_atom.waveform, '+:')
                plt.legend(('Signal', 'Atom substracted', 'residual'))
                plt.title('Iteration ' + str(it_number))
                dictionary.best_current_block.plot_proj_matrix()
                plt.show()

            raise ValueError()
        else:
            _Logger.warning("Something wrong happened at iteration %d ATTEMPT TO CONTINUE" % it_number)
            # attempt to continue after recomputing all projections
            dictionary.compute_touched_zone(best_atom, panic=True)
            SILENT_FAIL = False
            return _mp_loop(dictionary, debug, SILENT_FAIL, unpad,
                            res_signal, current_approx, res_energy, it_number)

#            return current_approx, res_energy
    if debug > 1:
        _Logger.debug(
            "new residual energy of " + str(np.sum(res_signal.data ** 2)))
    if not unpad:
        res_energy.append(res_signal.energy)
    else:
        padd = dictionary.get_pad()
        # assume padding is max dictionaty size
        res_energy.append(np.sum(res_signal.data[padd:-padd] ** 2))
        # only compute the energy without the padded borders where
    # eventually energy has been created
    # add atom to dictionary
    current_approx.add(best_atom)
    return best_atom


def _get_local_subdico(best_atom, current_approx, dictionary):
    # THIS IS CRITICAL STEP
    # The interval is defined so that any atom that overlaps the selected one
    # is selected
    t_interv = [int(best_atom.time_position - dictionary.get_pad()),
                int(best_atom.time_position + best_atom.length)]
    t_interv[0] = max(t_interv[0], 0)
    t_interv[1] = min(t_interv[1], current_approx.length -
                      dictionary.get_pad())
# ov_sel_atoms = current_approx.filter_atoms(time_interv=t_interv,
# index=True)
    ov_sel_atoms = current_approx.get_neighbors(best_atom)

    _Logger.debug(
        "%d neighbouring atoms selected in [%d - %d]" % (len(ov_sel_atoms),
                                                         t_interv[0],
                                                         t_interv[1]))

    # now build the subdictionary matrix
#    subdico = scipy.sparse.csr_matrix((t_interv[1] + dictionary.get_pad(
#    ) - t_interv[0], len(ov_sel_atoms) + 1))
    subdico = np.zeros((t_interv[1] + dictionary.get_pad(
    ) - t_interv[0], len(ov_sel_atoms) + 1))

    subdico[best_atom.time_position - t_interv[0]:
            best_atom.time_position - t_interv[0] + best_atom.length,
            -1] = _PyServer.get_waveform(best_atom.length, best_atom.freq_bin)

    for atomidx in range(len(ov_sel_atoms)):
        at = current_approx.atoms[ov_sel_atoms[atomidx]]

#        wf = at.get_waveform() / np.sqrt(np.sum(at.get_waveform() ** 2))

        subdico[
            at.time_position - t_interv[0]: at.time_position - t_interv[0] + at.length,
            atomidx] = _PyServer.get_waveform(at.length, at.freq_bin)

    return subdico, t_interv, ov_sel_atoms

# def _grad_proj(locproj, ov_sel_atoms, best_atom, t_interv, dict, current_approx):
#    """ avoids the numpy dot product, slow when numpy is not parallelized
#        also there are a lot of zeroes in the subdictionary, no need to take them
#        into account"""
#    # initialize output
#    outvec =  np.zeros((t_interv[1] + dict.get_pad() - t_interv[0], ))
#
#    # projection loop
#    for at_i in range(len(ov_sel_atoms)):
#        at = current_approx[ov_sel_atoms[at_i]]
#        wf = _PyServer.get_waveform( at.length, at.freq_bin)
#        slice = range(int(at.time_position - t_interv[0]),
#                      int(at.time_position - t_interv[0] + at.length))
#        outvec[slice] += locproj[at_i]*wf
#
#    # don't forget the last atom
#    wf = _PyServer.get_waveform( best_atom.length, best_atom.freq_bin)
#    slice = range(int(best_atom.time_position - t_interv[0]),
#                  int(best_atom.time_position - t_interv[0] + best_atom.length))
#    outvec[slice] += locproj[-1]*wf
#
#    return outvec


def _locomp_loop(dictionary, debug, silent_fail,
                 unpad, res_signal, current_approx,
                 res_energy, it_number):
    #===========================================================================
    # # Internal OMP loop with local least squares projection
    #=========================================================================

    # Compute inner products and selects the best atom
    # SAME AS MP
    dictionary.update(res_signal, it_number)
    # retrieve the best correlated atom
    best_atom = dictionary.get_best_atom(debug)
    if best_atom is None:
        _Logger.info('No atom selected anymore')

    if debug > 0:
        _Logger.debug(_itprint_(it_number, best_atom))

    # This is where things change a little:
    # first we need to retrieve all atoms previsously
    # selected that overlaps with the selected one
    subdico, t_interv, ov_sel_atoms = _get_local_subdico(
        best_atom, current_approx, dictionary)

    # recompute the weights by least squares on the residual signal
    weights = np.dot(np.linalg.pinv(subdico), res_signal.data[t_interv[
                     0]:t_interv[1] + dictionary.get_pad()])

    # update the approximant: change atom values
    current_approx.update(ov_sel_atoms, weights)

    # update the approximant waveform, locally
    current_approx.recomposed_signal.data[t_interv[0]:
                                          t_interv[1] + dictionary.get_pad()] += np.dot(subdico, weights)

    # TODO remove or optimize
    current_approx.recomposed_signal.energy = np.sum(
        current_approx.recomposed_signal.data ** 2)

    # add atom to dictionary: but not the waveform, already taken care of
    current_approx.add(best_atom, noWf=True)

    # update the residual, locally
    res_signal.data = current_approx.original_signal.data - \
        current_approx.recomposed_signal.data
    res_signal.energy = np.sum(res_signal.data ** 2)

    # compute the touched zone to accelerate further processings
    dictionary.compute_touched_zone(best_atom)

    if debug > 1:
        _Logger.debug(
            "new residual energy of " + str(np.sum(res_signal.data ** 2)))
    if not unpad:
        res_energy.append(res_signal.energy)
    else:
        padd = dictionary.get_pad()
        # assume padding is max dictionaty size
        res_energy.append(np.sum(res_signal.data[padd:-padd] ** 2))
        # only compute the energy without the padded borders where

    return best_atom


def _locgp_loop(dictionary, debug, silent_fail,
                unpad, res_signal, current_approx,
                res_energy, it_number):
    #===========================================================================
    # # Internal GP loop with local gradient updates
    #=========================================================================

    # Compute inner products and selects the best atom
    # SAME AS MP
    dictionary.update(res_signal, it_number)
    # retrieve the best correlated atom
    best_atom = dictionary.get_best_atom(debug)
    if best_atom is None:
        _Logger.info('No atom selected anymore')

    if debug > 0:
        _Logger.debug(_itprint_(it_number, best_atom))

    # Same as for locOMP : build neighbouring subdictionary
    # selected that overlaps with the selected one
    subdico, t_interv, ov_sel_atoms = _get_local_subdico(
        best_atom, current_approx, dictionary)

    # recompute the weights by least squares on the residual signal
    # Replace the pseudo-inverse by a gradient descent step
    if 'LO' in dictionary.nature:
        sigslice = res_signal.data[t_interv[0]:t_interv[1] +
                                   dictionary.get_pad()]

#         @TODO can we replace this step by a projection using fftw?
#         We don't have to do that, we already calculated them!!
# the projection score are all stored in the projection_matrix of the
# dictionary
        locProj = np.dot(subdico.T, sigslice)

    # HACK faster if weights have already been computed
    else:
        loc_projs2 = []
        for at_i in ov_sel_atoms:
            at = current_approx[at_i]
            block = dictionary.blocks[dictionary.find_block_by_scale(at.length)]
            print at.frame * at.length / 2 + at.freq_bin, at.freq_bin, at.frame, block.projs_matrix.shape
            loc_projs2.append(
                block.projs_matrix[at.frame * at.length / 2 + at.freq_bin])
        loc_projs2.append(best_atom.mdct_value)
        locProj = np.array(loc_projs2)

#    grad_proj = _grad_proj(locProj, ov_sel_atoms,
#                           best_atom, t_interv, dictionary, current_approx)
#    print subdico.shape, locProj.shape
    # projection

    grad_proj = np.dot(subdico, locProj)
    # refactoring: replace by a broadcasted computation since
    # all we need is the energy of the projected array
#    grad_proj_en = np.sum(np.sum((subdico*locProj),axis=1)**2)

#    print grad_proj_en - np.sum(grad_proj**2)

#    step = np.sum(locProj**2)/ np.sum(np.sum((subdico*locProj),axis=1)**2)
    step = np.sum(locProj ** 2) / np.sum(grad_proj ** 2)
#    print step
    # compute new weights
    weights = step * locProj

    # update the approximant: change atom values
    # @TODO optimize here: this is bringing serious overhead : DONE
    current_approx.update(ov_sel_atoms, weights)

    # update the approximant waveform, locally: np.dot(subdico,weights)
    # REFACTORED, no need to recompute dot product
    current_approx.recomposed_signal.data[t_interv[0]:
                                          t_interv[1] + dictionary.get_pad()] += step * grad_proj

    # @TODO remove or optimize
    current_approx.recomposed_signal.energy = np.sum(
        current_approx.recomposed_signal.data ** 2)

    # add atom to dictionary: but not the waveform, already taken care of
    current_approx.add(best_atom, noWf=True)

    # update the residual, locally
    res_signal.data = current_approx.original_signal.data - \
        current_approx.recomposed_signal.data
    res_signal.energy = np.sum(res_signal.data ** 2)

    # compute the touched zone to accelerate further processings
    dictionary.compute_touched_zone(best_atom)

    if debug > 1:
        _Logger.debug(
            "new residual energy of " + str(np.sum(res_signal.data ** 2)))
    if not unpad:
        res_energy.append(res_signal.energy)
    else:
        padd = dictionary.get_pad()
        # assume padding is max dictionaty size
        res_energy.append(np.sum(res_signal.data[padd:-padd] ** 2))
        # only compute the energy without the padded borders where

    return best_atom


def _omp_loop(dictionary, debug, silent_fail,
              unpad, res_signal, current_approx,
              res_energy, it_number):
    #===========================================================================
    # # Internal OMP loop with local gradient updates
    # BEWARE: VERY MEMORY CONSUMING!
    #=========================================================================

    # Compute inner products and selects the best atom
    # SAME AS MP
    dictionary.update(res_signal, it_number)
    # retrieve the best correlated atom
    best_atom = dictionary.get_best_atom(debug)
    if best_atom is None:
        _Logger.info('No atom selected anymore')

    if debug > 0:
        _Logger.debug(_itprint_(it_number, best_atom))

    if it_number == 0:
        dictionary.matrix = np.zeros((current_approx.length, 1))
    else:
        # add the current waveform to the subdictionary:
        dictionary.matrix = np.concatenate((dictionary.matrix,
                                            np.zeros((dictionary.matrix.shape[0], 1))), axis=1)

    t_interv = [int(best_atom.time_position - dictionary.get_pad()),
                int(best_atom.time_position + best_atom.length)]
    t_interv[0] = max(t_interv[0], 0)
    t_interv[1] = min(t_interv[1], current_approx.length -
                      dictionary.get_pad())
    # fill the new column with new waveform
    dictionary.matrix[best_atom.time_position:
                      best_atom.time_position +
                      best_atom.length, -1] = _PyServer.get_waveform(best_atom.length,
                                                                     best_atom.freq_bin)

    subdico = dictionary.matrix[:, 0:it_number + 1]

    # recompute the weights by least squares on the residual signal
#    print subdico.shape, res_signal.data.shape
    weights = np.dot(np.linalg.pinv(subdico), res_signal.data)

    # update the approximant: change atom values
    current_approx.update(range(it_number), weights)

    # update the approximant waveform, locally
    current_approx.recomposed_signal.data += np.dot(subdico, weights)

    # TODO remove or optimize
    current_approx.recomposed_signal.energy = np.sum(
        current_approx.recomposed_signal.data ** 2)

    # add atom to dictionary: but not the waveform, already taken care of
    current_approx.add(best_atom, noWf=True)

    # update the residual, locally
    res_signal.data = current_approx.original_signal.data - \
        current_approx.recomposed_signal.data
    res_signal.energy = np.sum(res_signal.data ** 2)

    # compute the touched zone to accelerate further processings
    dictionary.compute_touched_zone(best_atom)

    if debug > 1:
        _Logger.debug(
            "new residual energy of " + str(np.sum(res_signal.data ** 2)))
    if not unpad:
        res_energy.append(res_signal.energy)
    else:
        padd = dictionary.get_pad()
        # assume padding is max dictionaty size
        res_energy.append(np.sum(res_signal.data[padd:-padd] ** 2))
        # only compute the energy without the padded borders where

    return best_atom


def _initialize_fftw(dictionary, max_thread_num=None):
    if max_thread_num is not None:
        max_thread_flag = max_thread_num
    else:
        max_thread_flag = -1
    if parproj.initialize_plans(np.array(dictionary.sizes),
                                np.array([2] * len(dictionary.sizes)),
                                max_thread_flag) != 1:
        raise ValueError("Something failed during FFTW initialization step ")


def _clean_fftw():
    if parproj.clean_plans() != 1:
        raise ValueError("Something failed during FFTW cleaning stage ")


def mp_joint(orig_sig_list,
             dictionary,
             target_srr,
             max_it_num,
             debug=0,
             pad=True,
             escape=False,
             escape_threshold=0.4,
             escape_thr_map=None,
             bss_eval=False,
             sources=None,
             interval=100,
             clean=True,
             waitbar=True,
             no_adapt=False,
             silent_fail=False):
    """ Joint Matching Pursuit : Takes a bunch of signals in entry and decomposes the common part out of them
        Gives The common model and the sparse residual for each signal in return """

    # back compatibility - use debug levels now
    if debug is not None:
        _Logger.set_level(debug)
    _Logger.info("Call to mp.mp_joint")
    # We now work on a list of signals: we have a list of approx and residuals
    res_sig_list = []
    current_approx_list = []
    res_energy_list = []
    current_srr_list = []

    if escape:
        esc_approx_list = []
        esc_it_list = []
        criterions = []

#    if bss_eval:
#        SDRs = [];
#        SIRs = [];
#        SARs = [];
#    dictionaryList = []
    threshold_dict = {}
    if escape_thr_map is None:
        for size in dictionary.sizes:
            threshold_dict[size] = escape_threshold
    else:
        for size, value in zip(dictionary.sizes, escape_thr_map):
            threshold_dict[size] = value

    # create a mean approx of the background
    mean_approx = Approx.Approx(
        dictionary, [], orig_sig_list[0], debug_level=debug)
    k = 1
    for orig_signal in orig_sig_list:

        _Logger.debug("Initializing Signal Number " + str(k))

#        if pad:
#            orig_signal.pad(dictionary.get_pad())
        res_sig_list.append(orig_signal.copy())

        # initialize approximant
        current_approx_list.append(
            Approx.Approx(dictionary, [], orig_signal, debug_level=debug))

        if escape:
            esc_approx_list.append(Approx.Approx(dictionary, [
            ], orig_signal, debug_level=debug))
            esc_it_list.append([])
        # residualEnergy
        res_energy_list.append([])
        current_srr_list.append(current_approx_list[-1].compute_srr())
        k += 1

    # initialize blocks using the first signal: they shoudl all have the same
    # length
    _Logger.debug("Initializing Dictionary")
    dictionary.initialize(res_sig_list)

    # FFTW Optimization for C code: initialize module global variables
    try:
        if parproj.initialize_plans(np.array(dictionary.sizes), np.array(dictionary.tolerances)) != 1:
            raise ValueError(
                "Something failed during FFTW initialization step ")
    except:
        _Logger.error("Initialization step failed")
        raise

    iterationNumber = 0

    approxSRR = max(current_srr_list)

    # Decomposition loop: stopping criteria is either SNR or iteration number
    while (approxSRR < target_srr) & (iterationNumber < max_it_num):

        _Logger.info("mp LOOP : iteration " + str(iterationNumber + 1))
        # Compute inner products and selects the best atom
        dictionary.update(res_sig_list, iterationNumber)

        if debug > 0:
            maxScale = dictionary.best_current_block.scale
            maxFrameIdx = math.floor(
                dictionary.best_current_block.maxIdx / (0.5 * maxScale))
            maxBinIdx = dictionary.best_current_block.maxIdx - maxFrameIdx * (
                0.5 * maxScale)

            _Logger.debug("It: " + str(iterationNumber) + " Selected atom "
                          + str(dictionary.best_current_block.maxIdx)
                          + " of scale " + str(maxScale) + " frequency bin "
                          + str(maxBinIdx)
                          + " value : " + str(
                          dictionary.max_block_score)
                          + " Frame : " + str(maxFrameIdx))

        # retrieve the best correlated atoms, locally adapted to the signal
        best_atom_list = dictionary.get_best_atom(debug, noAdapt=no_adapt)

        if best_atom_list is None:
            print 'No atom selected anymore'
            raise ValueError('Failed to select an atom')

        escape_current_atom = False

        if escape:
            # Escape mechanism : if variance of amplitudes is too big : assign
            # it to the biggest only
            mean = np.mean([abs(atom.get_value()) for atom in best_atom_list])
            std = np.std([abs(atom.get_value()) for atom in best_atom_list])
            maxValue = np.max(
                [abs(atom.get_value()) for atom in best_atom_list])
            _Logger.debug("Mean : " + str(mean) + " - STD : " + str(std))

            criterions.append(std / mean)
#            print criterions[-1]
            if (std / mean) > threshold_dict[atom.length]:
                escape_current_atom = True
# print "Escaping Iteration ",iterationNumber,": mean " , mean , " std " , std
                _Logger.debug("Escaping!!")

        for sigIdx in range(len(res_sig_list)):

            if not escape_current_atom:
                # add atom to current regular approx
                current_approx_list[sigIdx].add(
                    best_atom_list[sigIdx], clean=False)

                dictionary.compute_touched_zone(sigIdx, best_atom_list[sigIdx])

                # subtract atom from residual
                try:
                    res_sig_list[
                        sigIdx].subtract(best_atom_list[sigIdx], debug)
                except ValueError:
                    if silent_fail:
                        continue
                    else:
                        raise ValueError("Subtraction of atom failed")
            else:
                # Add this atom to the escape approx only if this signal is a
                # maxima
                if abs(best_atom_list[sigIdx].get_value()) == maxValue:
#                if True:
#                    print "Added Atom to signal " + str(sigIdx)
                    esc_approx_list[sigIdx].add(best_atom_list[sigIdx])
                    current_approx_list[sigIdx].add(best_atom_list[sigIdx])
                    esc_it_list[sigIdx].append(iterationNumber)

                    # subtract atom from residual
                    res_sig_list[
                        sigIdx].subtract(best_atom_list[sigIdx], debug)

                    dictionary.compute_touched_zone(
                        sigIdx, best_atom_list[sigIdx])

                else:
                    _Logger.debug("Atom not subtracted in this signal")
                    dictionary.compute_touched_zone(
                        sigIdx, best_atom_list[sigIdx])

            # update energy decay curves
            res_energy_list[sigIdx].append(res_sig_list[sigIdx].energy)

            if debug > 0 or (iterationNumber % interval == 0):
                current_srr_list[sigIdx] = current_approx_list[
                    sigIdx].compute_srr(res_sig_list[sigIdx])

                _Logger.debug("Local adaptation of atom " + str(sigIdx) +
                              " - Position : " + str(best_atom_list[sigIdx].time_position) +
                              " Amplitude : " + str(best_atom_list[sigIdx].proj_score) +
                              " TimeShift : " + str(best_atom_list[sigIdx].time_shift))
#            if clean and sigIdx>0:
#                del best_atom_list[sigIdx].waveform;

        # also add the mean atom to the background model UNLESS this is an
        # escaped atom
        if not escape_current_atom:
#            mean_approx.add(best_atom_list[0] )
            mean_approx.add(
                dictionary.get_mean_atom(getFirstAtom=False), clean=clean)
            _Logger.debug("Atom added to common rep ")

        if clean:
            for sigIdx in range(len(res_sig_list)):
                del best_atom_list[sigIdx].waveform

#        dictionary.compute_touched_zone()

#        approxSRR = currentApprox.compute_srr();

        _Logger.debug("SRRs reached of " + str(current_srr_list) +
                      " at iteration " + str(iterationNumber))

#        if bss_eval and ( (iterationNumber+1) % interval ==0):
#            estimSources = np.zeros(sources.shape)
#            # first estim the source for the common part
#            estimSources[0, :,0] = mean_approx.recomposedSignal.dataVec
#
#            for sigIdx in range(len(res_sig_list)):
# estimSources[sigIdx+1, :,0] =
# current_approx_list[sigIdx].recomposedSignal.dataVec +
# esc_approx_list[sigIdx].recomposedSignal.dataVec;
# estimSources[sigIdx+1, :,0] =
# esc_approx_list[sigIdx].recomposedSignal.dataVec;
#            [SDR,ISR,SIR,SAR] = bss_eval_images_nosort(estimSources,sources)
#
##            print SDR , SIR
#            SDRs.append(SDR)
#            SIRs.append(SIR)
#            SARs.append(SAR)
#
##            print current_srr_list

        iterationNumber += 1
        if (iterationNumber % interval == 0):
            print iterationNumber
            print [resEnergy[-1] for resEnergy in res_energy_list]
    # VERY IMPORTANT CLEANING STAGE!
    if parproj.clean_plans(np.array(dictionary.sizes)) != 1:
        raise ValueError("Something failed during FFTW cleaning stage ")
#    if waitbar and (iterationNumber %(max_it_num/100) ==0):
#        print float(iterationNumber)/float(max_it_num/100) , "%",

    if not escape:
        return mean_approx, current_approx_list, res_energy_list, res_sig_list
    else:
        # time to add the escaped atoms to the corresponding residuals:
#        for sigIdx in range(len(res_sig_list)):
# res_sig_list[sigIdx].dataVec +=
# esc_approx_list[sigIdx].recomposedSignal.dataVec;
#        if bss_eval:
# return mean_approx, current_approx_list, res_energy_list , res_sig_list ,
# esc_approx_list , criterions , SDRs, SIRs , SARs
        return mean_approx, current_approx_list, res_energy_list, res_sig_list, esc_approx_list, esc_it_list



#*****************************************************************************/
#                                                                            */
#                           Tools.SoundFile.py                               */
#                                                                            */
#                        Matching Pursuit Library                            */
#                                                                            */
# M. Moussallam                                             Mon Aug 16 2010  */
# -------------------------------------------------------------------------- */

# A handle for audio files, including routines to Read, Write, Resample, Concatenate and Crop

class SoundFile:

    def Help(self):
        print """ python class SOUNDFILE
        Handles for reading, writing, resampling audio files in WAV format
        This utility manipulates multi-channel audio PCM signals as matrices
        """
    # properties
    data = None; # The data as a numpy array will be contained here

    # Constructor
    def __init__(self, filename):
        import wave, numpy
#        print filename
        if not len(filename) >0:
            print "Invalid wavfile name : %s"%filename
        
        elif filename[-3:] == 'raw':
            self.filename = filename[(filename.rfind('/'))+1:]
            self.filepath = filename[:(filename.rfind('/'))]
            wavfile = open(filename , 'r');

            self.filetype = filename[len(filename)-3:]
            self.nbChannel = 1 # fixed value
            self.sampleRate = 11025 # fixed value
            #self.nframes = wavfile.getnframes() # fixed value
            self.sample_width = 2 # format is signed int16
            #data = data/max([abs(min(data)) max(data)]);
            str_bytestream = wavfile.read(-1)
            self.data = numpy.fromstring(str_bytestream,'h')
            self.nframes = len(self.data)
            wavfile.close()

        elif filename[-3:] == 'wav' or filename[-3:] == 'WAV' or filename[-2:] == 'au':
            self.filename = filename[(filename.rfind('/'))+1:]
            self.filepath = filename[:(filename.rfind('/'))]
            
            if filename[-2:] == 'au':
                import sunau
                wavfile = sunau.open(filename , 'r');
            else:
                wavfile = wave.open(filename, 'r')
            self.filetype = filename[len(filename)-3:]
            self.nbChannel = wavfile.getnchannels()
            self.sampleRate = wavfile.getframerate()
            self.nframes = wavfile.getnframes()
            self.sample_width = wavfile.getsampwidth()

            #self.data = array.array('h') #creates an array of ints
            str_bytestream = wavfile.readframes(self.nframes)

            #print filename,self.sampleWidth, self.nbChannel , self.sampleRate,self.nframes

            if self.sample_width == 1:
                typeStr = 'int8'
            elif self.sample_width == 2:
                typeStr = 'int16'
            elif self.sample_width == 3:
                typeStr ='int24' # WARNING NOT SUPPORTED BY NUMPY
            elif self.sample_width == 4:
                typeStr = 'uint32'
            self.data = numpy.fromstring(str_bytestream,dtype=typeStr)

#            self.data = numpy.fromstring(str_bytestream,'h')
            wavfile.close()
        else:
            raise TypeError('Audio format not recognized')

    # ToString method
    def ToString(self):
        print self.filename , 'in ' , self.filepath
        #print sf.filetype
        print "Channels :" , self.nbChannel
        print "Sample Frequency :" , self.sampleRate , " Hz"
        print "Duration :" , (self.nframes), " frames  or " , float(self.nframes)/float(self.sampleRate) , ' seconds'
        print "Sample width :" , self.sample_width , ' bytes'

    # returns a numerical matrix of the file
    def GetAsMatrix(self):
        # nomalize?
        return self.data

    # writes to an output file
    def Write(self, destination, filename=''):
        import wave
        if not len(filename) >0:
            outputname = self.filename
        else:
            outputname = filename

        if not len(destination) >0:
            outputpath = self.filepath
        else:
            outputpath = destination

        wavfile = wave.open(outputpath + outputname, 'w')
        wavfile.setparams((self.nbChannel, self.sample_width , self.sampleRate , self.nframes , 'NONE', 'noncompressed'))
        wavfile.writeframes(self.data.tostring())
        wavfile.close()

        print 'File written to ' , outputpath + outputname




""" this module handles Modified Discrete Cosine Transform and Inverse transform

M. Moussallam as part of the PyMP package
"""
# from numpy import *


from numpy.fft import fft, ifft
from numpy import mod, zeros, array, sin, concatenate, real
from math import pi, sqrt
import cmath


def mdct(x, L):

    # Signal length
    N = len(x)

    # Number of frequency channels
    K = L / 2

    # Test length
    if mod(N, K) != 0:
        print "Input length must be a multiple of half window size."
        raise ValueError(N, " is not a multiple of ", K)

    # Pad edges with zeros
    x = concatenate((concatenate((zeros(L / 4), x)), zeros(L / 4).T))

    # Number of frames
    P = N / K
    if P < 2:
        print "Signal too short"
        raise ValueError

    # Framing
    temp = x.copy()
    output = zeros(N)
    x = zeros((L, P), complex)
    y = zeros((L, P), complex)

    # Windowing
    wLong = array([sin(float(l + 0.5) * (pi / L)) for l in range(L)])

    wEdgeL = wLong.copy()
    wEdgeL[0:L / 4 - 1] = 0
    wEdgeL[L / 4:L / 2 - 1] = 1
    wEdgeR = wLong.copy()
    wEdgeR[L / 2:L / 2 + L / 4 - 1] = 1
    wEdgeR[L / 2 + L / 4:L - 1] = 0

    # twidlle coefficients
    pre_twidVec = array([cmath.exp(n * (-1j) * pi / L) for n in range(L)])
    post_twidVec = array([cmath.exp(
        (float(n) + 0.5) * -1j * pi * (L / 2 + 1) / L) for n in range(L / 2)])

    # we're note in Matlab anymore, loops are more straight than indexing
    for i in range(P):
        x[:, i] = temp[i * K: i * K + L]
        if(i == 0):
            x[:, i] = x[:, i] * wEdgeL
        elif(i == P - 1):
            x[:, i] = x[:, i] * wEdgeR
        else:
            x[:, i] = x[:, i] * wLong

        # do the pre-twiddle
        x[:, i] = x[:, i] * pre_twidVec

        # compute fft
        y[:, i] = fft(x[:, i], L)

        # post-twiddle
        y[0:L / 2, i] = y[0:L / 2, i] * post_twidVec

#        output[i*K : (i+1)*K] = sqrt(2/float(K))*real(y[0:L/2 , i]);
        output[i * K: (i + 1) * K] = sqrt(2 / float(K)) * real(y[0:L / 2, i])

    return output

# inverse transform still using fft calculation


def imdct(y, L):
    # Signal length
    N = len(y)

    # Number of frequency channels
    K = L / 2

    # Test length
    if mod(N, K) != 0:
        print ["Input length ", N,
               " must be a multiple of half window size : ", K]
        raise ValueError(
            "The length was not a multipe of half the window size")

    # Pad edges with zeros
    # x = concatenate( (concatenate( (zeros(L/4) , x) )  , zeros(L/4).T) );

    # Number of frames
    P = N / K
    if P < 2:
        print "Signal too short"
        raise ValueError

    # Framing
    temp = y.copy()
    output = zeros(P * K + K)

    x = zeros((L, P), complex)
    y = zeros((L, P), complex)

    # Windowing
    wLong = array([sin(float(l + 0.5) * (pi / L)) for l in range(L)])
    wEdgeL = wLong.copy()
    wEdgeL[0:L / 4 - 1] = 0
    wEdgeL[L / 4:L / 2 - 1] = 1
    wEdgeR = wLong.copy()
    wEdgeR[L / 2:L / 2 + L / 4 - 1] = 1
    wEdgeR[L / 2 + L / 4:L - 1] = 0

    # twidlle coefficients
    pre_twidVec = array([cmath.exp(1j * 2 * float(
        n) * pi * (float(float(L / 4 + 0.5)) / float(L))) for n in range(L)])
    post_twidVec = array(
        [cmath.exp(0.5 * 1j * 2 * pi * (float(n) + (L / 4 + 0.5)) / L) for n in range(L)])

    # we're note in Matlab anymore, loops are more straight than indexing
    for i in range(P):
        y[0:K, i] = temp[i * K: (i + 1) * K]
        # do the pre-twiddle
        y[:, i] = y[:, i] * pre_twidVec

        # compute ifft
        x[:, i] = ifft(y[:, i])

        # post-twiddle
        x[:, i] = x[:, i] * post_twidVec

        # real part and scaling
#        x[:, i] = sqrt(2/float(K))*L*real(x[:, i]);
        x[:, i] = 2 * sqrt(1 / float(L)) * L * real(x[:, i])
        if(i == 0):
            x[:, i] = x[:, i] * wEdgeL
        elif(i == P - 1):
            x[:, i] = x[:, i] * wEdgeR
        else:
            x[:, i] = x[:, i] * wLong

        # overlapp - add
        output[i * K: i * K + L] = output[i * K: i * K + L] + x[:, i]

    # scrap zeroes on the borders
    return output[K / 2:-K / 2]



#
#                                                                            */
#                               signals.py                                   */
#                                                                            */
#                        Matching Pursuit Library                            */
#                                                                            */
# M. Moussallam                                             Mon Aug 16 2010  */
# -------------------------------------------------------------------------- */
#

import math
import wave
#import scikits.audiolab as audiolab
import struct
import numpy as np

from tools import SoundFile
import base
import log

global _Logger
_Logger = log.Log('Signal', imode=False)
# from operator import add, mul


class Signal(base.BaseSignal):
    """This file defines the main class handling audio signal in the Pymp Framework.

    A Signal is fairly a numpy array (called data) and a collection of attributes.
    Longer signals should not be loaded in memory. the `LongSignal` class allows
    to define this kind of signal, and slice it in frames overlapping or not.

    Attributes
    ----------
    *data* : array
        a numpy array containing the signal data
    *channel_num* : int
        The number of channel
    *length* : int
        The length in samples (integer) of the signal (total dimension of the numpy array is channel_num x length)
    *fs* : int
        The sampling frequency
    *location* :  str , optional
        Where the original file is located on the disk
    *sample_width* : int
        Various bit format exist for wav file, this allows to handle it
    *is_normalized* : bool
        A boolean telling is the numpy array has been normalized (which here means its values are between -1 and 1)
    *energy* : float
        The energy (:math:`\sum_i x[i]^2`) of the array
    


    Notes
    -----
    For use in the PyMP framework, other methods need to be defined:
        :func:`add` :             adds an atom waveform to the current signal and updates energy
        :func:`subtract` :        the opposite operation
    

    """

    # att
    data = np.array([])
    channel_num = 0
    length = 0
    fs = 0
    location = ""
    sample_width = 2
    is_normalized = False
    energy = 0

    # Constructor
    def __init__(self, data=[], Fs=0, normalize=False, mono=False, debug_level=None,
                 fast_create=False):
        """ Simple constructor from a numpy array (data) or a string 
    parameters
    ----------
    data : array-like
        a numpy array containing the signal data
    Fs :  int
        the sampling frequency
    normalize : bool, optionnal
        whether to normalize the signal (e.g. max will be 1)
    mono :  bool, optionnal
        keep only the first channel
    fast_create : bool, optionnal
        for fast creation: skip test on the input data
        """
        if fast_create:                            
            # fast creation : not computing the energy nor checking input format
            self.data = data
            self.length = self.data.shape[0]
            self.fs = Fs
        else:
            if isinstance(data, str):
                self.location = data
                try:
                    import scikits.audiolab as audiolab
                    Sf = audiolab.Sndfile(data, 'r')
                    self.data = Sf.read_frames(int(Sf.nframes))
                    Fs = Sf.sampleRate
                    self.sample_width = Sf.encoding
                except:
                    Sf = SoundFile.SoundFile(data)
                    self.data = Sf.GetAsMatrix().reshape(Sf.nframes, Sf.nbChannel)
                    self.sample_width = Sf.sample_width
                    Fs = Sf.sampleRate
                
            else:        
                self.data = np.array(data)
                # remove any nans or infs data
                self.data[np.isnan(self.data)] = 0
                self.data[np.isinf(self.data)] = 0
    
                if len(self.data.shape) > 2:
                    raise ValueError("Cannot process more than 2D arrays")
    
                if len(self.data.shape) > 1:
                    if self.data.shape[0] < self.data.shape[1]:
                        # by convention we store channels as columns...
                        self.data = self.data.transpose()
    
                    
            # keeping only the left channel
            if mono and len(self.data.shape) > 1:
                self.data = self.data[:, 0]
    
            if debug_level is not None:
                _Logger.set_level(debug_level)
    
            self.length = self.data.shape[0]
            self.fs = Fs
    
            if len(self.data.shape) > 1:
                self.channel_num = self.data.shape[1]
            else:
                self.channel_num = 1
    
            if normalize & (self.length > 0):
                _Logger.info('Normalizing Signal')
                normaFact = self.data.max()
                self.data = self.data.astype(float) / float(normaFact)
                self.is_normalized = True
    
            if len(data) > 0:
                self.energy = np.sum(self.data ** 2)
    
            _Logger.info('Signal created ' + str(self))

    def __getitem__(self, item):
        if isinstance(item, slice):
            return Signal(self.data[item], self.fs, normalize=False)
        else:
            raise TypeError("Argument not recognized as a slice")

    def __repr__(self):
        return '''Signal object located in %s
        length: %d
        energy of %2.3f
        sampling frequency %d
        number of channels %d''' % (self.location, self.length, self.energy, self.fs, self.channel_num)

    def normalize(self):
        ''' makes sure all values of the array are between -1 and 1
            CAUTION: this is an amplitude normalization not an energy one'''
        _Logger.info('Normalizing Signal')
        normaFact = abs(self.data).max()
        self.data = self.data.astype(float) / float(normaFact)
        self.energy = sum(self.data ** 2)
        self.is_normalized = True

    def plot(self, pltStr='b-', legend=None):
        ''' DEPRECATED plot the array using matplotlib '''
        import matplotlib.pyplot as plt
        plt.plot(self.data, pltStr)
        if legend != None:
            plt.legend((legend))
        plt.show()

    def play(self, prevent_too_long=True, int_type=np.int16):
        '''EXPERIMENTAL: routine to play the signal using wave and pyaudio'''
        import pyaudio
        p = pyaudio.PyAudio()

        # TODO: allow other types of integers
#        p.get_format_from_width(2)
        stream = p.open(format=8,
                        channels=self.channel_num,
                        rate=self.fs,
                        output=True)

        # data = wf.readframes(CHUNK)
        # Warning: check whether the flow has been normalized
        # in which case we need to remultiply it

        if self.is_normalized:
            data = (self.data * 16384).astype(int_type).tostring()
        else:
            data = self.data.astype(int_type).tostring()

        # TODO : check the size of the input
        if (self.length / self.fs > 10) and prevent_too_long:
            raise ValueError("File is longer than 10 secs, please call with prevent_too_long=False")
        stream.write(data)

        stream.stop_stream()
        stream.close()
        p.terminate()

    # cropping routine
    def crop(self, startIndex=0, stopIndex=None):
        ''' cropping routine, usage is quite obvious '''
        if (startIndex > self.length) | (startIndex < 0):
#            raise ValueError("Starting Index beyond signal dimensions")
            _Logger.error('Starting Index ' + str(
                startIndex) + ' is beyond signal length of ' + str(startIndex))
        if (stopIndex > self.length) | (stopIndex < 0):
            _Logger.error('Stopping Index ' + str(
                stopIndex) + ' is beyond signal length of ' + str(self.length))
#            raise ValueError("Stopping Index beyond signal dimensions")
        if (len(self.data.shape) > 1):
            if (self.data.shape[0] > self.data.shape[1]):
                self.data = self.data[startIndex:stopIndex, :]
            else:
                self.data = self.data[:, startIndex:stopIndex]
        else:
            self.data = self.data[startIndex:stopIndex]
        self.length = stopIndex - startIndex

    def window(self, K):
        """ apply a sine window on norders """
        if K > 0:
            self.data[0:K] *= np.sin((np.arange(K).astype(float)
                                      ) * np.pi / (2 * K))
            self.data[-K:] *= np.sin(
                (np.arange(K).astype(float)) * np.pi / (2 * K) + np.pi / 2)

    def copy(self):
        copiedSignal = Signal(self.data.copy(), self.fs)
        copiedSignal.location = self.location
        copiedSignal.channel_num = self.channel_num
        copiedSignal.sample_width = self.sample_width
        copiedSignal.is_normalized = self.is_normalized
        return copiedSignal

    def subtract(self, atom, debug=0, prevent_energy_increase=True):
        ''' Subtracts the atom waveform from the signal, at the position specified by the atom.timeLocalization property
            if prevent_energy_increase is True, an error will be raised if subtracting the atom increases the signal's energy '''
        if not isinstance(atom, base.BaseAtom):
            raise TypeError("argument provided is not an atom")
        if atom.waveform is None:
#            print "Resynth"
            atom.synthesize()
        # Keep track of the energy to make
        oldEnergy = (self.data[atom.time_position: atom.
                               time_position + atom.length] ** 2).sum()

        # In high debugging levels, show the process using matplotlib
        if debug > 2:
            import matplotlib.pyplot as plt
            plt.figure()
            plt.plot(self.data[atom.time_position: atom.
                               time_position + atom.length])
            plt.plot(atom.waveform)
            plt.plot(self.data[atom.time_position: atom.
                               time_position + atom.length] - atom.waveform, ':')
            plt.legend(('Signal', 'Atom substracted', 'residual'))
            plt.show()

        # actual subtraction
        self.data[atom.time_position: atom.time_position + atom.
                  length] -= atom.waveform

        newEnergy = (self.data[atom.time_position: atom.
                               time_position + atom.length] ** 2).sum()

        # Test
        if (newEnergy > oldEnergy) and prevent_energy_increase:
            # do not substract the atom
            self.data[atom.time_position: atom.
                      time_position + atom.length] += atom.waveform
            # get out of here
            err_str = """"Warning : Substracted Atom created energy:
                          from %1.6f to from %1.6f :
%s""" % (oldEnergy, newEnergy, atom)
            _Logger.error(err_str)
            raise ValueError('see Logger')

        self.energy = self.energy - oldEnergy + newEnergy

    def resample(self, newFs):
        """ resampling the signal """
        from scipy.signal import resample
        resamp_data = resample(self.data, self.get_duration()*newFs)
        self.data = resamp_data
        self.length = len(self.data)
        self.fs = newFs
    
    def downsample(self, newFs):
        """ downsampling the signal by taking only a portion of the data """

        if newFs >= self.fs:
            _Logger.warning('WARNING new sampling frequency is bigger than actual, trying upsampling instead')
            return self.resample(newFs)

        # ratio of selected points
        subRatio = float(self.fs) / float(newFs)
        indexes = np.floor(np.arange(0, self.length, subRatio))
#        print indexes
        self.data = self.data[indexes.tolist()]
        self.length = len(self.data)
        self.fs = newFs

    def add(self, atom):
        ''' adds the contribution of the atom at the position specified by the atom.timeLocalization property '''
        if not isinstance(atom, base.BaseAtom):
            raise TypeError("argument provided is not an atom")

        if atom.waveform is None:
            _Logger.info("Resynthesizing  waveform")
            atom.synthesize()

        localEnergy = np.sum(self.data[atom.time_position: atom.
                                       time_position + atom.length] ** 2)

        # do sum calculation
        try:
            self.data[atom.time_position: atom.
                      time_position + atom.length] += atom.waveform
        except:
            _Logger.error('Mispositionned atom: from %d to %d - sig length of %d'%(atom.time_position,
                                                                atom.length,
                                                                self.length))
#        # update energy value
        self.energy += np.sum(self.data[atom.time_position: atom.
                                        time_position + atom.length] ** 2) - localEnergy

    # Pad edges with zeroes
    def pad(self, zero_pad):
        ''' Pad edges with zeroes '''
        try:
            self.data = np.concatenate((np.concatenate(
                (np.zeros(zero_pad), self.data)), np.zeros(zero_pad)))
            self.length = len(self.data)
        except ValueError:
            print(self.data.shape)
            _Logger.error(
                "Wrong dimension number %s" % str(self.data.shape))

    def get_duration(self):
        ''' returns the duration in seconds '''
        return float(self.length) / float(self.fs)

    def depad(self, zero_pad):
        ''' Remove zeroes from the edges WARNING: no test on the deleted data: make sure these are zeroes! '''
        try:
            self.data = self.data[zero_pad:-zero_pad]
            self.length = len(self.data)
        except ValueError:
            print(self.data.shape)
            _Logger.error(
                "Wrong dimension number %s" % str(self.data.shape))

#    def doWindow(self, K):
#        ''' DEPRECATED '''
#        self.data[0:K] *= np.sin((np.arange(K).astype(float)) * np.pi / (2 * K))
#        self.data[-K:] *= np.sin((np.arange(K).astype(float)) * np.pi / (
#            2 * K) + np.pi / 2)

    def write(self, fileOutputPath, normalize=False, pad=0):
        ''' Write the current signal at the specified location in wav format

            This is done using the wave python library'''
        if pad > 0:
            self.depad(pad)

        if self.energy == 0:
            _Logger.warning("Zero-energy signal")
            self.data = np.zeros(self.length,)
#            print "Warning!!!! Zero-energy signal:"

        wav_file = wave.open(fileOutputPath, 'wb')
        # BUGFIX
        if isinstance(self.sample_width, str):
            self.sample_width = 2
        if self.channel_num < 1:
            self.channel_num = 1
        wav_file.setparams((self.channel_num,
                            self.sample_width,
                            self.fs,
                            self.length,
                            'NONE', 'not compressed'))

        # prepare binary output
        values = []
        if not self.is_normalized and normalize:
            self.normalize()
        clipped = False
        
        for i in range(len(self.data)):
            if np.isnan(self.data[i]):
#                _Logger.warning("NaN data found: replaced by 0")
                self.data[i] = 0
            if self.data[i] < -1:
#                _Logger.warning(str(i) + 'th sample was below -1: cropping')
                self.data[i] = -1
            if self.data[i] > 1:                
#                _Logger.warning(str(i) + 'th sample was over 1: cropping')
                self.data[i] = 1
                clipped=True
            value = int(16384 * self.data[i])
            packed_value = struct.pack('h', value)
            values.append(packed_value)

        if clipped:
            _Logger.warning('Some values clipped during the writing.. \
                         you may want to normalize to get values -1 and 1')
        value_str = ''.join(values)
        wav_file.writeframes(value_str)
        import os
        _Logger.info("Written to %s"% os.path.abspath(fileOutputPath))
#        else:
# _Logger.warning("Warning BUGFIX not done yet, chances are writing failed");
#            wav_file.writeframes((self.data).astype(int).tostring())
        wav_file.close()

    def wigner_plot(self, window=True):
        """ Calculate the wigner ville distribution and plots it
        WARNING: not sufficiently tested!"""
        from numpy.fft import ifft, fft, ifftshift, fftshift

        N = self.length
        Ex = self.data

        # ensure the signal is mono (take left canal if stereo)
        if len(Ex.shape) > 1 and not (Ex.shape[1] == 1):
            Ex = Ex[:, 0]

        # reshape anyway as a column vector
        Ex = Ex.reshape((N, 1))

        if window:
            from scipy.signal import hann
            Ex *= hann(N).reshape((N, 1))

        x = ifftshift((np.arange(0, N) - N / 2) * 2 * np.pi / (N - 1))
        #%   Generate linear vector
        X = np.arange(0, N) - N / 2

        x = x.reshape((N, 1))
        X = X.reshape((1, N))

        A = np.dot(1j * x, X / 2)
        EX1 = ifft(np.dot(fft(Ex), np.ones((1, N))) * np.exp(A))
        #%   +ve shift
        EX2 = ifft(
            np.dot(fft(Ex), np.ones((1, N))) * np.exp(np.dot(-1j * x, X / 2)))
        #%   -ve shift

        tmp0 = EX1 * np.conj(EX2)

        print

        print tmp0.shape
        tmp1 = fftshift(tmp0, (1, ))
        tmp2 = fft(tmp1, axis=1)

        W = np.real(fftshift(tmp2, (1,)))
        #%   Wigner function

        return W

    def spectrogram(self, wsize=512, tstep=256, order=2, log=False,
                    ax=None, cmap=None, cbar=True):
        """ Compute and plot the (absolute value) spectrogram of the signal 
        
        Based on a short-time fourier transform
        
        Parameters
        ----------
        wsize : int
            length of the STFT window in samples (must be a multiple of 4)
        tstep : int , opt
            step between successive windows in samples (must be a multiple of 2,
            a divider of wsize and smaller than wsize/2) (default: wsize/2)
        order : int, opt
            1 will have the Magnitude Spectrum, 2 the Power Spectrum
        log : bool, opt
            Set to True for the log-spectrogram
            
        returns
        -------
        
        The Spectrogram matrix
        """
        import matplotlib.pyplot as plt
        from scipy.fftpack import fft, ifft, fftfreq
        if ax is None:            
            ax = plt.gca()
                    
        
        ### Errors and warnings ###
        if wsize % 4:
            raise ValueError('The window length must be a multiple of 4.')
    
        if tstep is None:
            tstep = wsize / 2
    
        tstep = int(tstep)
    
        if (wsize % tstep) or (tstep % 2):
            raise ValueError('The step size is %d but it should be a multiple of 2 and a '
                             'divider of the window length %d'%(tstep,wsize))
    
        if tstep > wsize:
            raise ValueError('The step size is %d but it should be smaller than the '
                             'window length %d'%(tstep,wsize))
    
        n_step = int(math.ceil(self.length / float(tstep)))
        n_freq = wsize / 2 + 1
        
        _Logger.debug("Number of frequencies: %d" % n_freq)
        _Logger.debug("Number of time steps: %d" % n_step)
    
        X = np.zeros((self.channel_num, n_freq, n_step), dtype=np.complex)
    
        if self.channel_num == 0:
            return X
        x = self.data.T
        T = self.length
        # Defining sine window
        win = np.sin(np.arange(.5, wsize + .5) / wsize * np.pi)
        win2 = win ** 2
    
        swin = np.zeros((n_step - 1) * tstep + wsize)
        for t in range(n_step):
            swin[t * tstep:t * tstep + wsize] += win2
        swin = np.sqrt(wsize * swin)
    
        # Zero-padding and Pre-processing for edges
        xp = np.zeros((self.channel_num, wsize + (n_step - 1) * tstep),
                      dtype=x.dtype)
        xp[:, (wsize - tstep) / 2: (wsize - tstep) / 2 + T] = x
        x = xp
    
        for t in range(n_step):
            # Framing
            wwin = win / swin[t * tstep: t * tstep + wsize]
            frame = x[:, t * tstep: t * tstep + wsize] * wwin[None, :]
            # FFT
            fframe = fft(frame)
            X[:, :, t] = fframe[:, :n_freq]
        
        Spectro = np.abs(X)**order
        
        if log:
            Spectro = np.log10(Spectro)
        
        if Spectro.shape[0]>1:
            print "Taking mean of channels"
            Spectro = np.squeeze(np.mean(Spectro, axis=0))
        else:
            Spectro = np.squeeze(Spectro)
        
        (Fmax,Tmax) = Spectro.shape
        yticks =  np.arange(.0,Fmax, Fmax/10.0)
        yvalues = (yticks*0.5*float(self.fs)/float(Fmax)).astype(int)
        xticks =  np.arange(.0,Tmax, Tmax/10.0)
        xvalues = (xticks*float(tstep)/float(self.fs))
        if cmap is None:
            import matplotlib.cm as cm
            cmap = cm.coolwarm
        plt.imshow(Spectro,
                   origin='lower',
                   cmap=cmap)
        
        plt.xticks(xticks, ["%1.2f"%v for v in xvalues])
        plt.yticks(yticks, yvalues)
        plt.xlabel('Time (s)')
        plt.ylabel('Frequency (Hz)')
        if cbar:
            plt.colorbar()
            
        return Spectro
# def InitFromFile(filepath, forceMono=False, doNormalize=False, debugLevel=None):
#    ''' Static method to create a Signal from a wav file on the disk
#        This is based on the wave Python library through the use of the Tools.SoundFile class
#        '''
#    if debugLevel is not None:
#        _Logger.set_level(debugLevel)
#    Sf = SoundFile.SoundFile(filepath)
#    #print Sf.GetAsMatrix().shape
#    reshapedData = Sf.GetAsMatrix().reshape(Sf.nframes, Sf.nbChannel)
#    if forceMono:
#        reshapedData = reshapedData[:, 0]
#
#    sig = Signal(reshapedData, Sf.sampleRate, doNormalize)
#    sig.sample_width = Sf.sample_width
#    sig.location = filepath
#
#    _Logger.info("Created Signal of length " + str(
#        sig.length) + " samples of " + str(sig.channel_num) + "channels")
#    # print "Created Signal of length " + str(Signal.length) +" samples " #of "
#    # + str(Signal.channel_num) + "channels"
#    return sig


class LongSignal(Signal):
    """ A class handling long audio signals

        Subclass of :class:`.Signal` where the data is not loaded at once for memory consumptions purposes
        Instead, the data is sliced in frames that can be loaded later individually. very useful to
        process large files such as audio archives

        Attributes
        ----------
        `filepath`:    The path to the audio file
        `frame_size`:   In samples, defaut is 16384*3
        `frame_duration`:  Alternative to frameSize, specify directly a frame duration in seconds (Defult is None)
        `mono`:      Only load the left (first) channel (default is False)
        `Noverlap` :      overlap (as a ratio r such that :math:`0\leq r < 1`)

    Standard methods to manipulate the signal are:

        :func:`get_sub_signal` : Loads a subSignal

    """

    # more attr
#    filetype = ''       # file extension : .wav or .mp3
#    nframes = 0;        # number of
#    sample_width = 0;    # bit width of each frame
# segment_size = 0;      # in 16-bits samples : different from nframes which is
# internal wav representation
#    n_seg =0;     # numberof audio segments to be consideres
#    overlap = 0;         # overlapping rate between two consecutive segments

    # constructor
    def __init__(self, filepath, frame_size=16384 * 3, frame_duration=None, mono=False, Noverlap=0):
        self.location = filepath
        self.segment_size = frame_size

        try:
            import scikits.audiolab as audiolab
            wavfile = audiolab.Sndfile(filepath, 'r')
            self.sample_width = wavfile.encoding
            self.channel_num = wavfile.channels
            self.fs = wavfile.samplerate
            self.nframes = wavfile.nframes
        except:
            print "audiolab not installed"
            wavfile = None
            wavfile = wave.open(filepath, 'r')
        # overlaps methods from SoundFile object
        
#        if (filepath[-4:] =='.wav'):
#            wavfile = wave.open(filepath, 'r')
#        elif (filepath[-3:] =='.au'): 
#            import sunau
#            wavfile = sunau.open(filepath , 'r');
#        elif (filepath[-4:] =='.raw'):
#            wavfile = open()
            self.sample_width = wavfile.getsampwidth()            
            self.channel_num = wavfile.getnchannels()
            self.fs = wavfile.getframerate()
            self.nframes = wavfile.getnframes()
        
        self.filetype = filepath[len(filepath) - 3:]
        self.overlap = Noverlap

        if frame_duration is not None:
            # optionally set the length in seconds, adapt to the sigbal
            # sampling frequency
            self.segment_size = int(math.floor(
                frame_duration * self.fs))

        if not (self.nframes % self.segment_size) == 0:
            _Logger.warning("Not a round number of segment: cropping ")
            self.nframes = (
                self.nframes / self.segment_size) * self.segment_size

        self.n_seg = int(math.floor(
            float(self.nframes) / (float(self.segment_size) * (1.0 - self.overlap))))

        if self.overlap >= 1:
            raise ValueError('Overlap must be in [0..1[ ')

#        print self.nframes, (float(self.segment_size) * (1.0 - self.overlap))
        self.n_seg -= int(math.ceil(self.overlap / (1.0 - self.overlap)))

        _Logger.info('Loaded ' + filepath + ' , ' + str(
            self.nframes) + ' frames of ' + str(self.sample_width) + ' bytes')
        _Logger.info('Type is ' + self.filetype + ' , ' + str(self.
                                                              channel_num) + ' channels at ' + str(self.fs))
        _Logger.info('Separated in ' + str(self.n_seg) + ' segments of size ' + str(self.segment_size) + ' samples overlap of ' + str(self.overlap * self.segment_size))
        self.length = self.n_seg * (1.0- Noverlap)*frame_size
        wavfile.close()

#    def readFrames(self , frameIndexes):
#        #self.data = array.array('h') #creates an array of ints
#        wavfile = wave.open(self.location, 'r')
#        str_bytestream = wavfile.readframes(self.nframes)
#        self.data = fromstring(str_bytestream,'h')
#        wavfile.close()

    def get_sub_signal(self, start_seg, seg_num, mono=False, normalize=False,
                       channel=0, pad=0, fast_create=False):
        """ Routine to actually read from the buffer and return a smaller signal instance

        :Returns:

            a :class:`Signal` object

        :Example:

            longSig = LongSignal(**myLongSigFilePath**, frameDuration=5) # Initialize long signal
            subSig = longSig.get_sub_signal(0,10) # Loads the first 50 seconds of data

        """

        # convert frame into bytes positions
        bFrame = int(start_seg * (self.segment_size * (1 - self.overlap)))
        nFrames = int(seg_num * self.segment_size)
        
        try:
            import scikits.audiolab as audiolab            
            wavfile = audiolab.Sndfile(self.location, 'r')
            wavfile.seek(bFrame, 0, 'r')
            data = wavfile.read_frames(nFrames)
        except:
            wavfile = wave.open(self.location, 'r')
        
#        print "Reading ",bFrame, nFrames, wavfile._framesize
            str_bytestream = wavfile.read_frames(nFrames)            
            wavfile.close()

        if self.channel_num > 1:
            reshapedData = data.reshape(nFrames, self.channel_num)
        else:
            if max(data.shape) > nFrames:
                nFrames = sum(data.shape)

            reshapedData = data.reshape(nFrames, )

        if mono:
            if len(reshapedData.shape) > 1:
                reshapedData = reshapedData[:, channel]

            reshapedData = reshapedData.reshape(nFrames, )

        SubSignal = Signal(reshapedData, self.fs,
                           normalize=normalize,
                           fast_create=fast_create)
        SubSignal.location = self.location

        if pad != 0:
            SubSignal.pad(pad)

# print "Created Signal of length " + str(SubSignal.length) +" samples " #of "
# + str(Signal.channel_num) + "channels"
        return SubSignal


    def get_duration(self):
        ''' returns the duration in seconds '''
        return float(self.length) / float(self.fs)